# -*- coding: utf-8; py-indent-offset:4 -*-

import requests
import pandas as pd

_EXCHANGE_LIST = ['nyse', 'nasdaq', 'amex']

_REGIION_LIST = ['AFRICA','EUROPE','ASIA','AUSTRALIA+AND+SOUTH+PACIFIC','CARIBBEAN','SOUTH+AMERICA','MIDDLE+EAST','NORTH+AMERICA']

_SECTORS_LIST = ['Consumer Non-Durables', 'Capital Goods', 'Health Care',
       'Energy', 'Technology', 'Basic Industries', 'Finance',
       'Consumer Services', 'Public Utilities', 'Miscellaneous',
       'Consumer Durables', 'Transportation']

# headers and params used to bypass NASDAQ's anti-scraping mechanism in function __exchange2df
nasdaq_headers = {
    'authority': 'api.nasdaq.com',
    'accept': 'application/json, text/plain, */*',
    'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36',
    'origin': 'https://www.nasdaq.com',
    'sec-fetch-site': 'same-site',
    'sec-fetch-mode': 'cors',
    'sec-fetch-dest': 'empty',
    'referer': 'https://www.nasdaq.com/',
    'accept-language': 'en-US,en;q=0.9',
}

def stocks_params(exchange):
    return (
        ('letter', '0'),
        ('exchange', exchange),
        ('download', 'true'),
    )

#
# Data source: exchange = nyse, nasdaq, amex
# https://api.nasdaq.com/api/screener/stocks?exchange=nasdaq&download=true
#
def download_us_stock_list(param = None, verbose = False):
    if param is not None:
        print('\rfetching data from nasdaq ...', end = '', flush = True)
        r = requests.get('https://api.nasdaq.com/api/screener/stocks', headers=nasdaq_headers, params=stocks_params(param))
        print('\r', end = '', flush = True)
        data = r.json()['data']
        df = pd.DataFrame(data['rows'], columns=data['headers'])
    else:
        df = pd.DataFrame()
        for exchange in _EXCHANGE_LIST:
            ex_df = download_us_stock_list(exchange, verbose)
            df = df.append(ex_df, ignore_index=True)
    # removes weird tickers
    df = df[['symbol','name','country','ipoyear']]
    df = df[~df['symbol'].str.contains(r'\.|\^', regex=True)]
    df = df.sort_values(by='symbol').reset_index(drop=True)
    if verbose:
        print(df)
    return df

def index_params(offset):
    return (
        ('letter', '0'),
        ('offset', offset),
        ('download', 'true'),
    )

#
# Data source:
# https://api.nasdaq.com/api/screener/index?offset=0&limit=50
#
def download_us_index_list(param = None, verbose = False):
    df = None
    offset = 0
    for i in range(10):
        print('\rfetching data from nasdaq ...', end = '', flush = True)
        r = requests.get('https://api.nasdaq.com/api/screener/index', headers=nasdaq_headers, params=index_params(offset))
        print('\r', end = '', flush = True)
        records = r.json()['data']['records']
        totalRecords = int(records['totalrecords'])
        data = records['data']
        if df is None:
            df = pd.DataFrame(data['rows'], columns=data['headers'])
        else:
            df = df.append( pd.DataFrame(data['rows'], columns=data['headers']) , ignore_index= True)
        offset = int(records['offset']) + int(records['limit'])
        if offset > totalRecords:
            break
    df.columns = ['symbol', 'name', 'price', 'netchange', 'pctchange']
    df = df[['symbol','name']]
    df = df.sort_values(by='symbol').reset_index(drop=True)
    if verbose:
        print(df)
    return df




# -*- coding: utf-8 -*-

from .basic import CROSS
from .ma import SMA

from ..core.indicator_signal import register_signal_indicator
# Different of Moving Average, by
# --------------------------------------------
def DMA(close, fast = 10, slow = 50, signal = 10):
    dma = SMA(close, fast) - SMA(close, slow)
    ama = SMA(dma, signal)
    return dma, ama

def signal_dma(df, inplace=False, fast=10, slow=50, signal=10):
    dma, ama = DMA(df.close, fast=fast, slow=slow, signal=signal)
    if inplace:
        df['dma'], df['dma_ama'] = dma, ama
    return CROSS(dma, ama)

register_signal_indicator('dma', signal_dma, ['dma', 'dma_ama'], 'DMA', 'trend')



# -*- coding: utf-8 -*-
from .basic import *

from .bias import *
from .boll import *
from .cci import *
from .dma import *
from .ene import *
from .kdj import *
from .ma import *
from .macd import *
from .mfi import *
from .obv import *
from .rsi import *
from .sar import *
from .trix import *
from .vhf import *
from .wr import *
from .mffi import *



# -*- coding: utf-8; py-indent-offset:4 -*-

import multiprocessing as mp

from ..utils import date_from_str
from .stock_market import Market
from .trader import Trader
from .fund import Fund

def backtest_one(args):
    strategy, start_cash, date_start, date_end = args

    market = Market(date_start, date_end)
    trader = Trader(market)

    fund = Fund(market, trader)
    fund.set_name(strategy.__name__)
    fund.set_start_cash( start_cash )
    fund.add_strategy( strategy(fund) )
    trader.add_fund(fund)
    trader.run_fund(date_start, date_end)

    return trader.get_report()

def backtest_strategy(strategy, start_cash= None, date_start= None, date_end= None, compare_index= None, out_file= None, parallel= True):
    if type(strategy) != list:
        strategy = [ strategy ]

    if start_cash is None:
        start_cash = 1000000.00
    if date_start is None:
        date_start = date_from_str('3 years ago')
    if date_end is None:
        date_end = date_from_str('yesterday')

    if parallel:
        args_list = [[stra, start_cash, date_start, date_end] for stra in strategy]
        with mp.Pool(len(strategy)) as p:
            report_list = p.map(backtest_one, args_list)
        all_report = []
        for report in report_list:
            all_report += report
        trader = Trader(Market(date_start, date_end))

    else:
        market = Market(date_start, date_end)
        trader = Trader(market)
        for stra in strategy:
            fund = Fund(market, trader)
            fund.set_name(stra.__name__)
            fund.set_start_cash( start_cash )
            fund.add_strategy( stra(fund) )
            trader.add_fund(fund)
        trader.run_fund(date_start, date_end)
        all_report = trader.get_report()

    trader.print_report(all_report)
    trader.plot(report = all_report, compare_index= compare_index, out_file= out_file)



# -*- coding: utf-8; py-indent-offset:4 -*-
from .datetime import date_from_str, datetime_today

def parse_params_options(argv):
    params = []
    options = []
    for i in range(1, len(argv)):
        str = argv[i]
        if str[0] == '-':
            options.append(str)
        else:
            params.append(str)

    return params, options

def parse_arg(argv, required, defaults, syntax_tips):
    params, options = parse_params_options(argv)

    n = len(params)
    if n < required:
        print(syntax_tips.replace('__argv0__', argv[0]))
        exit()

    vals = list(defaults.values())
    m = len(vals)
    if n < m:
        params = params + vals[n:]

    return params[0:m], options

def dict_from_config_items(items, verbose = False):
    mapping = {}
    for k, v in items:
        mapping[ k ] = v
        if verbose:
            print(k, '=', v)
    return mapping

def date_limit_from_options(options):
    limit = 0
    date_from = None
    date_to = None
    days = None
    for option in options:
        if option.startswith('-days='):
            days = int(option.replace('-days=',''))
            date_from = date_from_str('{} days ago'.format(days))
            date_to = datetime_today()
        if option.startswith('-date='):
            date_range = option.replace('-date=','').split('-')
            date_from = date_from_str(date_range[0])
            date_to = date_from_str(date_range[1]) if (len(date_range)>1 and len(date_range[1])>0) else datetime_today()
        if option.startswith('-limit='):
            limit = int(option.replace('-limit=',''))

    if date_from is None:
        days = 365 * 1
        date_from = date_from_str('{} days ago'.format(days))
        date_to = datetime_today()

    return date_from, date_to, limit

def csv_xlsx_from_options(options):
    csv = ''
    xlsx = ''
    for k in options:
        if k.startswith('-out='):
            if k.endswith('.csv'):
                csv = k.replace('-out=', '')
            if k.endswith('.xlsx'):
                xlsx = k.replace('-out=', '')
    return csv, xlsx




# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import importlib.util
import pandas as pd
import datetime as dt
import matplotlib.pyplot as plt

from ..utils import earn_to_annual
from .order_cost import get_order_cost, OrderCost
from .agent_simulated import SimulatedAgent
from .lang import LANG
from .data_cache import get_all_symbol_name, get_symbol_name

class Fund:
    market = None
    trader = None
    conf = None
    fund_name = 'No Name'
    strategy_list = None
    agent = None
    stat_df = None

    verbose = False

    stat_file = None
    plot_file = None
    compare_index = None

    def __init__(self, market, trader, fund_conf = None):
        self.market = market
        self.trader = trader
        self.conf = fund_conf if (fund_conf is not None) else {}
        self.strategy_list = []
        self.agent = None
        self.stat_df = None
        self.fund_name = ''
        self.start_cash = 0.0
        self.set_agent(SimulatedAgent(market))

    def set_verbose(self, verbose = True):
        self.verbose = verbose
        if self.agent is not None:
            self.agent.set_verbose(verbose)

    def add_strategy(self, strategy):
        if strategy not in self.strategy_list:
            self.strategy_list.append( strategy )
        pass

    def set_agent(self, agent, order_cost = None):
        self.agent = agent
        self.agent.order_cost = order_cost if (order_cost is not None) else get_order_cost()

    def set_stat_file(self, stat_file):
        self.stat_file = stat_file

    def set_plot_file(self, plot_file):
        self.plot_file = plot_file

    def set_compare_index(self, compare_index):
        self.compare_index = compare_index

    def load_module(self, strategy_file):
        mod_name = os.path.basename(strategy_file).replace('.py', '')
        mod_filepath = os.path.abspath(strategy_file)
        if self.verbose:
            print('dynamically loading strategy:', mod_filepath)
        spec = importlib.util.spec_from_file_location(mod_name, mod_filepath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def load_strategy(self, strategy_file):
        mod = self.load_module( strategy_file )
        mod.init( self )

    def set_name(self, name):
        self.fund_name = name

    def set_start_cash(self, cash):
        self.start_cash = cash

    def init_fund(self, realtime_trade = False):
        if 'name' in self.conf:
            self.fund_name = self.conf['name']
        if 'start_cash' in self.conf:
            self.start_cash = float(self.conf['start_cash'])

        if self.agent is None:
            self.set_agent(SimulatedAgent(self.market), OrderCost())

        # for backtrade, we force to use simulated agent
        if (not realtime_trade) and (self.agent.agent_type != 'simulated'):
            self.set_agent(SimulatedAgent(self.market), self.agent.order_cost)

        self.agent.init_portfolio( self.start_cash )

        self.max_value = self.agent.get_portfolio().total_value()

        columns = ['buy', 'sell', 'value', 'cash', 'position', 'drawdown']
        self.stat_df = pd.DataFrame([], columns = columns, index = pd.to_datetime([]))
        self.stat_df.index.name = 'date'
        self.update_stat(self.market.current_date - dt.timedelta(days=1))

        if 'strategy' in self.conf:
            self.load_strategy( self.conf['strategy'] )

    def update_stat(self, date = None):
        # calculate current value of today, save to dataframe
        agent = self.agent
        portfolio = agent.get_portfolio()
        value = portfolio.total_value()
        self.max_value = max(value, self.max_value)
        buy_count, sell_count = agent.get_transaction_count()
        row = [
            buy_count,
            - sell_count,
            value,
            portfolio.available_cash,
            (value - portfolio.available_cash) / value,
            (value / self.max_value -1),
        ]

        if date is None:
            date = self.market.current_date

        self.stat_df.loc[ pd.to_datetime(date) ] = pd.Series(row, self.stat_df.columns)

        if self.stat_file is not None:
            df = self.stat_df
            df.to_csv(self.stat_file, index= True)

        if (self.stat_df.shape[0] > 1) and (self.plot_file is not None):
            self.plot(out_file= self.plot_file)

    def before_day(self):
        self.agent.before_day()

    def before_tick(self):
        self.agent.before_tick()

    def after_tick(self):
        self.agent.after_tick()

    def after_day(self):
        self.agent.after_day()
        self.update_stat()

    def get_summary(self):
        df = self.stat_df
        if self.verbose:
            print(df)

        start_value = df.value.iloc[0]
        current_value = df.value.iloc[-1]
        earn = current_value / start_value -1

        days = (df.index[-1] - df.index[0]).days
        years = days / 365
        annual_earn = earn_to_annual(earn, days)

        return {
            'fund_name': self.fund_name,
            'start value': '{:,}'.format(round(start_value, 2)),
            'current value': '{:,}'.format(round(current_value, 2)),
            'invest year': round(years, 1),
            'trade count': int(sum(df.buy) - sum(df.sell)),
            'total earn': str(round(100 * earn, 2)) + ' %',
            'annual earn': str(round(100 * annual_earn, 2)) + ' %',
            'max drawdown': str(round(100 * min(df.drawdown), 2)) + ' %',
        }

    def get_stat(self):
        return self.stat_df

    def get_report(self):
        return {
            'name': self.fund_name,
            'summary': self.get_summary(),
            'stat': self.stat_df,
        }

    def print_report(self, report = None):
        if report is None:
            report = self.get_report()

        if type(report) != list:
            report = [ report ]

        summary_df = pd.DataFrame()
        for fund in report:
            symmary = fund['summary']
            if summary_df.shape[0] == 0:
                summary_df = pd.DataFrame([], columns = list(symmary.keys()))
            summary_df = summary_df.append(symmary, ignore_index = True)

        summary_df = summary_df.T
        summary_df.index.name = ''
        summary_df.columns.name = ''
        summary_df.index = summary_df.index.str.pad(15, side='left')

        print('-' * 80)
        print(summary_df)
        print('-' * 80)

    def plot(self, report = None, compare_index = None, out_file= None):
        if report is None:
            report = self.get_report()

        if type(report) != list:
            report = [ report ]

        dates = report[0]['stat'].index
        date_start = min(dates)
        date_end = max(dates)

        df = pd.DataFrame([], index=pd.date_range(start=date_start, end=date_end))
        for fund in report:
            fund_name = fund['name']
            stat_df = fund['stat']

            value = stat_df['value']
            df[ fund_name + '.' ] = (value / value.iloc[0] - 1) * 100.0
            df[ fund_name + ':'] = stat_df['drawdown'] * 100
            df[ fund_name + ','] = stat_df['position'] * 100
            df.index = stat_df.index

            # if only one strategy, we also plot the buy/sell
            if len(report) == 1:
                df['buy'] = stat_df['buy']
                df['sell'] = stat_df['sell']

        if compare_index is None:
            compare_index = self.compare_index

        if compare_index:
            symbol_name = get_all_symbol_name()
            compare_index_name = symbol_name[ compare_index ] if (compare_index in symbol_name) else get_symbol_name(compare_index)

            cmp_value = self.market.get_index_daily(compare_index, start=date_start, end=date_end)['close']
            cmp_value = (cmp_value / cmp_value.iloc[0] -1) * 100.0

            df[ compare_index_name ] = cmp_value

        df.dropna(inplace=True)
        df.index = df.index.strftime('%Y-%m-%d')

        plt.rcParams['font.sans-serif'] = ['SimHei'] # Chinese font
        plt.rcParams['font.family'] = 'sans-serif'
        plt.rcParams["axes.unicode_minus"] = False

        if len(report) > 1:
            plot_rows = 3
            grid_ratios = [3, 1, 1]
        else:
            plot_rows = 4
            grid_ratios = [3, 1, 1, 1]

        fig, axes = plt.subplots(nrows = plot_rows, gridspec_kw = {'height_ratios': grid_ratios})

        fund_list = []
        pos_list = []
        drawdown_list = []
        for k in df.columns:
            if k.endswith('.'):
                fund_list.append(k)
            elif k.endswith(':'):
                drawdown_list.append(k)
            elif k.endswith(','):
                pos_list.append(k)

        fund_list.append( compare_index_name )
        df[ fund_list ].plot(ax=axes[0], figsize = (10,6), grid = True, sharex=axes[0], label = 'date', ylabel = LANG('return(%)'), title = LANG('quantitative backtesting'))

        if len(report) == 1:
            ax_id = plot_rows -3
            axes[ax_id].set_ylabel(LANG('trade'))
            axes[ax_id].bar(df.index, df.buy, color='r')
            axes[ax_id].bar(df.index, df.sell, color='g')

        df[ pos_list ].plot(ax=axes[plot_rows -2], grid = True, sharex=axes[0], ylabel = LANG('position(%)'), legend=False)
        df[ drawdown_list ].plot(ax=axes[plot_rows -1], grid = True, sharex=axes[0], ylabel = LANG('drawdown(%)'), legend=False)

        #plt.xticks(rotation=30)

        if out_file is not None:
            plt.savefig(out_file)
        else:
            plt.show()



# -*- coding: utf-8; py-indent-offset:4 -*-

import time
import datetime as dt
import pandas as pd
import akshare as ak

def download_cn_index_daily( symbol ):
    daily_df = ak.stock_zh_index_daily_tx(symbol = symbol)
    daily_df['date'] = pd.to_datetime(daily_df.index.date)
    daily_df['volume'] = daily_df['amount']
    daily_df.set_index('date', inplace=True, drop=True)
    return daily_df



# -*- coding: utf-8; py-indent-offset:4 -*-

import pandas as pd
import akshare as ak

_fund_flow_common_column_mapping = {
    '主力净流入-净占比': 'main_pct',
    '超大单净流入-净占比': 'super_pct',
    '大单净流入-净占比': 'large_pct',
    '中单净流入-净占比': 'medium_pct',
    '小单净流入-净占比': 'small_pct',
    '主力净流入-净额': 'main_fund',
    #'超大单净流入-净额': 'super_fund',
    #'大单净流入-净额': 'large_fund',
    #'中单净流入-净额': 'medium_fund',
    #'小单净流入-净额': 'small_fund',
}
_fund_flow_days = {
    '1d': '今日',
    '3d': '3日',
    '5d': '5日',
    '10d': '10日',
}
_fund_flow_sectors = {
    'region': '地域资金流',
    'industry': '行业资金流',
    'concept': '概念资金流',
}

def list_fund_flow_days():
    return _fund_flow_days.keys()

def list_fund_flow_sectors():
    return _fund_flow_sectors.keys()

#
# Data source:
# http://data.eastmoney.com/zjlx/dpzjlx.html
#
def download_cn_market_fund_flow() -> pd.DataFrame:
    df = ak.stock_market_fund_flow()

    cols = {
        '日期': 'date',
        '上证-收盘价': 'sh_close',
        '深证-收盘价': 'sz_close',
        '上证-涨跌幅': 'sh_change_pct',
        '深证-涨跌幅': 'sz_change_pct',
    }
    cols.update(_fund_flow_common_column_mapping)
    df = df.rename(columns= cols)

    df['date'] = pd.to_datetime(df['date'])
    df = df.set_index('date', drop= True).astype('float64')

    selected_columns = list(cols.values())
    selected_columns.remove('date')
    df = df[ selected_columns ]
    for col in df.columns:
        if '_fund' in col:
            df[col] = df[col] / 100000000.00
    return df

#
# Data source:
# http://data.eastmoney.com/bkzj/hy.html
#
def download_cn_sector_fund_flow_rank(days= '1d', sector= 'industry') -> pd.DataFrame:
    if days not in _fund_flow_days:
        raise ValueError('invalid days: ' + days + ', must be one of: ' + ','.join(_fund_flow_days))
    if sector not in _fund_flow_sectors:
        raise ValueError('invalid sector: ' + sector + ', must be one of: ' + ','.join(_fund_flow_sectors))

    zh_days = _fund_flow_days[ days ]
    zh_sector = _fund_flow_sectors[ sector ]
    df = ak.stock_sector_fund_flow_rank(indicator= zh_days, sector_type= zh_sector)

    df.columns = df.columns.str.replace(zh_days, '')
    cols = {
        #'序号': 'index',
        '名称': 'sector',
        '涨跌幅': 'change_pct',
        '主力净流入最大股': 'top_stock',
    }
    cols.update(_fund_flow_common_column_mapping)
    df = df.rename(columns= cols)

    selected_columns = list(cols.values())
    df = df[ selected_columns ]

    for col in df.columns:
        if '_fund' in col:
            df[col] = df[col] / 100000000.00

    return df

#
# Data source:
# http://data.eastmoney.com/zjlx/detail.html
#
def download_cn_stock_fund_flow_rank(days = '1d') -> pd.DataFrame:
    if days not in _fund_flow_days:
        raise ValueError('invalid days: ' + days + ', must be one of: ' + ','.join(_fund_flow_days))
    zh_days = _fund_flow_days[ days ]

    print('\rfetching fund flow data ...', end = '', flush = True)
    df = ak.stock_individual_fund_flow_rank(indicator= zh_days)
    print('\r', end = '', flush = True)

    cols = {
        #'序号': 'index',
        '代码': 'symbol',
        '名称': 'name',
        '最新价': 'price',
        '涨跌幅': 'change_pct',
    }
    cols.update(_fund_flow_common_column_mapping)
    df.columns = df.columns.str.replace(zh_days, '')
    df = df.rename(columns= cols)

    selected_columns = list(cols.values())
    df = df[ selected_columns ]

    # convert number from str to float
    float_types = {}
    for col in df.columns:
        if col.endswith('price') or col.endswith('_fund') or col.endswith('_pct'):
            float_types[col] = 'float64'
            # some cells are '-', change to '0' to avoid error of type converting
            df[col] = ['0' if ((type(v) == str) and (v == '-')) else v for v in df[col]]
    df = df.astype( float_types )

    # number of is too large, we divide to unit '亿' and easier to read
    for col in df.columns:
        if '_fund' in col:
            df[col] = df[col] / 100000000.00

    return df

#
# Data source:
# http://data.eastmoney.com/zjlx/300750.html
#
def download_cn_stock_fund_flow(symbol) -> pd.DataFrame:
    market = 'sz' if (symbol[0] in ['0','3']) else 'sh'
    df = ak.stock_individual_fund_flow(symbol, market)

    cols = {
        '日期': 'date',
    }
    cols.update(_fund_flow_common_column_mapping)
    df = df.rename(columns= cols)

    df['date'] = pd.to_datetime(df['date'])
    df = df.set_index('date', drop= True).astype('float64')

    selected_columns = list(cols.values())
    selected_columns.remove('date')
    df = df[ selected_columns ]
    for col in df.columns:
        if '_fund' in col:

            df[col] = df[col] / 100000000.00
    return df

def download_cn_fund_list(param= None, verbose= False):
    df = ak.fund_em_open_fund_daily()
    return df

def download_cn_fund_info(symbol):
    df = ak.fund_em_open_fund_info(symbol, indicator='单位净值走势')
    return df

def download_cn_etf_fund_list(param= None, verbose= False):
    df = ak.fund_em_etf_fund_daily()
    return df

def download_cn_etf_fund_info(symbol):
    df = ak.fund_em_etf_fund_info(symbol)
    return df

def download_cn_fund_manager(param= None, verbose= False):
    df = ak.fund_manager(explode= True)
    return df



# -*- coding: utf-8 -*-

from .basic import CROSS, NON_ZERO, STDEV
from .ma import SMA
from ..core.indicator_signal import register_signal_indicator

# Bollinger Bands, by John Bollinger, 1980s
# --------------------------------------------
def BOLL(close, length = 20, std = 2.0, ddof=0):
    std = float(std) if std and std > 0 else 2.0
    ddof = int(ddof) if ddof >= 0 and ddof < length else 1
    stdev = STDEV(close, length, ddof)
    dev = std * stdev
    mid = SMA(close, length)
    lower = mid - dev
    upper = mid + dev
    bandwidth = (upper - lower) / NON_ZERO(mid) * 100.0
    return mid, upper, lower, bandwidth

def signal_boll(df, inplace=False, length=20, std=2, ddof=0):
    mb, up, dn, width = BOLL(df.close, length=length, std=std, ddof=ddof)
    if inplace:
        df['boll'], df['boll_upper'], df['boll_lower'], df['boll_width'] = mb, up, dn, width
    buy_signal = CROSS(df.close, dn) > 0
    sell_signal = CROSS(df.close, up) < 0
    return buy_signal.astype(int) - sell_signal.astype(int)

register_signal_indicator('boll', signal_boll, ['boll', 'boll_upper', 'boll_lower'], 'BOLL', 'obos')



# -*- coding: utf-8; py-indent-offset:4 -*-
import os
import sys
import datetime as dt
import pandas as pd
import tabulate as tb

from ..core.data_cache import get_cn_index_list_df, get_finance_indicator_df, update_finance_indicator_df, get_pepb_symmary_df
from ..utils import sort_with_options, filter_with_options

def cli_finance_help():
    syntax_tips = '''Syntax:
    __argv0__ finance <action> <symbols | stockpool.csv | all> [options]

Action:
    update ......................... download finance reports and update finance indicators
    view ........................... view finance indicators with filters

Symbols:
    <symbols> ...................... stock symbol list, like 600036 000002
    <stockpool.csv> ................ stock pool csv file
    all ............................ all symbols

Options:
    -sortby=<col> .................. sort by the column

    -ipo_years=<from>-<to> ......... IPO years between <from> to <to>
    -earn_ttm=<from>-<to> .......... annual earn between <from> to <to>
    -roe=<from>-<to> ............... ROE between <from> to <to>
    -grow_rate=<from>-<to> ......... average yearly grow rate between <from> to <to>
    -3yr_grow_rate=<from>-<to> ..... recent 3 year grow rate between <from> to <to>

    -tab ........................... show data in table format

    -out=<out.csv> ................. export selected stocks into <out.csv> file

Example:
    __argv0__ finance update 600036 000002 600276
    __argv0__ finance update my-stocks.csv
    __argv0__ finance update all

    __argv0__ finance view 600036 000002 600276
    __argv0__ finance view my-stocks.csv
    __argv0__ finance view all -ipo_years=1- -earn_ttm=1.0- -roe=0.15- -3yr_grow_rate=0.15- -sortby=roe -out=good_stock.csv
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print(syntax_tips)

def cli_finance_update(params, options):
    if len(params) == 0:
        cli_finance_help()
        return
    elif params[0] == 'all':
        symbols = get_cn_index_list_df()['symbol'].tolist()
    else:
        if params[0].endswith('.csv'):
            stock_df = pd.read_csv(params[0], dtype=str)
            symbols = stock_df['symbol'].tolist()
        else:
            symbols = params
    print(symbols)

    df = update_finance_indicator_df(symbols)
    print(df)

def cli_finance_view(params, options):
    df = get_finance_indicator_df()

    if (len(params) == 0) or (params[0] == 'all'):
        pass
    else:
        if params[0].endswith('.csv'):
            stock_df = pd.read_csv(params[0], dtype=str)
            symbols = stock_df['symbol'].tolist()
        else:
            symbols = params
        df = df[ df['symbol'].isin(symbols) ]
        df = df.set_index('symbol', drop= True)

        df['earn_speed'] = df['earn_speed'].round(3)
        df['earn_yoy'] = df['earn_yoy'].round(3)

        if '-pepb' in options:
            df_pepb = get_pepb_symmary_df(symbols)[['symbol','pe','pb','pe_pos','pb_pos']]
            df_pepb = df_pepb.set_index('symbol', drop= True)
            df = pd.concat([df, df_pepb], axis=1)
            df['peg'] = (df['pe'] / df['earn_yoy'] * 0.01).round(2)
            df['pe'] = df['pe'].round(1)
            df['pb'] = df['pb'].round(1)
            df.insert(0, 'symbol', df.index)
            df = df.reset_index(drop= True)

    total_n = df.shape[0]

    # now filter
    df = filter_with_options(df, options)
    filtered_n = df.shape[0]

    # now sort
    df = sort_with_options(df, options, by_default='roe')

    if '-tab' in options:
        print( tb.tabulate(df, headers='keys') )
    else:
        print('-' * 80)
        print(df)
        print('-' * 80)


    print('{} out of {} records selected.'.format(filtered_n, total_n))

    out_xlsx_file = ''
    out_csv_file = ''
    for k in options:
        if k.startswith('-out='):
            if k.endswith('.xlsx'):
                out_xlsx_file = k.replace('-out=', '')
            if k.endswith('.csv'):
                out_csv_file = k.replace('-out=', '')

    if out_csv_file:
        df_stocks = df[['symbol', 'name']]
        df_stocks.to_csv(out_csv_file, index= False)
        print('Exported to:', out_csv_file)
        print(df)

    if out_xlsx_file:
        name_col = df.pop('name')
        df.insert(1, 'name', name_col)
        df = df.drop(columns=['grow', 'share_grow', 'start_bvps', 'now_bvps', 'eps','assets','equity','shares'])
        df = df.rename(columns={
            'symbol':'代码',
            'name':'名称',
            'ipo_date':'上市日期',
            'ipo_years': '上市年数',
            'last_report': '最新财报',
            '3yr_grow_rate': '近3年增长率',
            'grow_rate': '平均增长率',
            'roe': 'ROE',
            '3yr_roe': '近3年ROE',
            'avg_roe': '平均ROE',
            'earn_speed': '利润环比增速',
            'earn_yoy': '利润同比增速',
            'debt_ratio': '资产负债率',
            'cash_ratio': '现金占比',
            'earn_ttm': '盈利TTM',
            'pe': 'PE',
            'pb': 'PB',
            'pe_pos': 'PE百分位',
            'pb_pos': 'PB百分位',
            'peg': 'PEG',
        })
        print( tb.tabulate(df, headers='keys') )
        df.to_excel(out_xlsx_file)
        print('Exported to:', out_xlsx_file)

    print('')

def cli_finance(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_finance_help()
        return

    action = params[0]
    params = params[1:]

    if action not in ['update', 'view', 'show', 'filter']:
        print('\nError: invalid action: ', action)
        return

    if action == 'update':
        cli_finance_update(params, options)
        return
    elif action in ['view', 'show', 'filter']:
        cli_finance_view(params, options)
    else:
        print('invalid action:', action)
        cli_finance_help()



# -*- coding: utf-8 -*-

import pandas as pd

def signal_no_dup(signal):
    sig = signal[0]
    for i in range(1, signal.size):
        s = signal.iloc[i]
        if s == 0:
            continue
        elif s == sig: # same
            signal.iloc[i] = 0
        else: # different, change compare target
            sig = s

    return signal

def signal_to_long(signal, time_factor = 1.0):
    # remove dup signal to avoid cumsum exceed 1
    signal = signal_no_dup(signal)

    # remove first trade signal if it's sell signal
    for i in range(signal.size):
        s = signal.iloc[i]
        if s == 0:
            continue
        elif s > 0:
            break
        else:
            signal.iloc[i] = 0
            break

    # cumsum, valid pos value should be either 0 or 1
    long_pos = signal.cumsum().fillna(0)

    if time_factor < 0 or time_factor > 1.0:
        raise ValueError('please make sure 0 <= long_factor <= 1')

    # if buy/sell in time, we can earn more and lose less
    long_factor = 1.0 - time_factor
    short_factor = time_factor
    pos = 0
    for i in range(long_pos.size):
        p = long_pos.iloc[i]
        if p > pos: # first long pos, it's a buy
            long_pos.iloc[i] = long_factor
            pos = p
        elif p < pos: # first short pos, it's a sell
            long_pos.iloc[i] = short_factor
            pos = p

    return long_pos

_registered_signal_indicators = {
    # id: {
    # 'func': function(),
    # 'cols': [],
    # 'type': '',
    # 'label': '',
    # }
}

def get_all_signal_indicators():
    return _registered_signal_indicators

def list_signal_indicators():
    return _registered_signal_indicators.keys()

def register_signal_indicator(id, func, cols, label, ind_type):
    _registered_signal_indicators[ id ] = {
        'func': func,
        'cols': cols,
        'type': ind_type,
        'label': label,
    }

def gen_indicator_signal(df, indicators, time_factor = 1.0, inplace=False):
    signal = pd.Series(0, index=df.index)
    for k in indicators:
        if k in _registered_signal_indicators:
            values = _registered_signal_indicators[k]
            signal_func = values['func']
            # use + to merge the signals
            signal += signal_func(df, inplace=inplace)
            if len(indicators) == 1:
                return signal
        else:
            print(k, 'not found in registered indicators')

    signal[ signal > 0 ] = 1
    signal[ signal < 0 ] = -1

    return signal_no_dup(signal)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
import datetime as dt
import configparser
import multiprocessing as mp

from ..utils import datetime_today, dict_from_config_items, seconds_from_str, date_from_str
from ..core import get_order_cost
from ..core import Market, Trader, Fund, EmailPush, SimulatedAgent, HumanAgent

def cli_trade_help():
    syntax_tips = '''Syntax:
    __argv0__ trade <action> <my-fund.conf> [options]

<my-fund.conf> ............ config file for one or multiple funds

<action>
    create ................ create a config file from template
    backtrade ............. backtrade test during a past date period
    run ................... run realtime from today to future

<options>
    -m .................... use multi-processing to boost backtrade
    -q .................... run a shorter backtrade

Example:
    __argv0__ trade create etc/myfund.conf
    __argv0__ trade backtrade etc/myfund.conf 20180101 20210101
    __argv0__ trade run etc/myfund.conf

Alias:
    __argv0__ backtrade etc/myfund.conf 20180101 20210101
    __argv0__ run etc/myfund.conf
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print(syntax_tips)

def get_fund_conf_template():
    return '''
[main]
tick_period = 5 min
compare_index = ^GSPC

[fund_list]
1 = fund_1

[fund_1]
name = My Fun No.1
start_cash = 1000000.00
agent = agent_1
strategy = strategy/mystrategy.py

[agent_1]
agent_type = human
portfolio_load = data/account_1_portfolio_load.csv
portfolio_save = data/account_1_portfolio_save.csv
order = data/account_1_order.csv
push_to = email_1

[email_1]
push_type = email
mailto = your_name@gmail.com
sender = no-reply@gmail.com
server = 192.168.0.200
user = 
passwd =
'''

def cli_trade_create(params, options):
    template_content = get_fund_conf_template()

    file_to_create = params[0]
    fp = open(file_to_create, 'w')
    fp.write(template_content)
    fp.close()

    print('-' * 80)
    print( template_content )
    print('-' * 80)

    print( 'Fund configuration file created:\n  ', os.path.abspath(file_to_create))
    print( '\nPlease edit file content before running.' )

def cli_run_fund(config_file, start, end, options):
    verbose = '-d' in options

    # read config file
    print( 'reading config from from:', config_file)
    if not os.path.isfile(config_file):
        print('Error:', config_file, 'not eixsts\n')
        exit()

    global_config = configparser.ConfigParser()
    global_config.read(config_file, encoding='utf-8')

    main_conf = dict_from_config_items(global_config.items('main'), verbose= verbose)
    compare_index = main_conf['compare_index'] if ('compare_index' in main_conf) else None

    date_start = date_from_str( start )
    date_end = date_from_str( end )

    start_tick = dt.datetime.now()

    # market is a global singleton
    market = Market(date_start - dt.timedelta(days=90), date_end)
    trader = Trader(market)

    for k, fund_id in global_config.items('fund_list'):
        fund_conf = dict_from_config_items(global_config.items(fund_id), verbose= verbose)
        fund = Fund(market, trader, fund_conf)

        if date_end > datetime_today():
            if 'stat_file' in fund_conf:
                fund.set_stat_file(fund_conf['stat_file'])

            if 'plot_file' in fund_conf:
                fund.set_plot_file(fund_conf['plot_file'])
                fund.set_compare_index( compare_index )

        if 'agent' in fund_conf:
            agent_conf = dict_from_config_items(global_config.items(fund_conf['agent']))
            agent_type = agent_conf['agent_type']
            if agent_type == 'human':
                agent = HumanAgent(market, agent_conf)
            #elif agent_type == 'automated':
            #    agent = AutomatedAgent(market, agent_conf)
            else:
                agent = SimulatedAgent(market, agent_conf)

            if (agent is not None) and ('push_to' in agent_conf):
                push_list = agent_conf['push_to'].replace(' ','').split(',')
                for push_to in push_list:
                    push_conf = dict_from_config_items(global_config.items(push_to))
                    push_type = push_conf['push_type']
                    if push_type == 'email':
                        agent.add_push_service(EmailPush(push_conf))
        else:
            agent = SimulatedAgent(market, {})

        fund.set_agent( agent, get_order_cost(global_config) )

        trader.add_fund(fund)

    market.set_verbose( verbose )
    trader.set_verbose( verbose )

    tick_period = seconds_from_str( main_conf['tick_period'] ) if ('tick_period' in main_conf) else 60
    trader.run_fund(date_start, date_end, tick_period = tick_period)

    trader.print_report()

    end_tick = dt.datetime.now()
    print('time used:', (end_tick - start_tick))

    out_file = None
    for option in options:
        if option.startswith('-out=') and option.endswith('.png'):
            out_file = option.replace('-out=', '')

    # compare with an index
    trader.plot(compare_index= compare_index, out_file= out_file)

    print('Done.\n')

# this func is prepared for multiprocessing
def cli_one_trader_run_one_fund(args):
    global_config, fund_id, date_start, date_end, tick_period, verbose = args

    start_tick = dt.datetime.now()

    fund_conf = dict_from_config_items(global_config.items(fund_id), verbose= verbose)

    market = Market(date_start, date_end)
    trader = Trader(market)
    fund = Fund(market, trader, fund_conf)

    agent = None
    if (date_end > datetime_today()) and ('agent' in fund_conf):
        agent_conf = dict_from_config_items(global_config.items(fund_conf['agent']))
        agent_type = agent_conf['agent_type']
        if agent_type == 'human':
            agent = HumanAgent(market, agent_conf)
        #elif agent_type == 'automated':
        #    agent = AutomatedAgent(market, agent_conf)
        else:
            agent = SimulatedAgent(market, agent_conf)

        if (agent is not None) and ('push_to' in agent_conf):
            push_list = agent_conf['push_to'].replace(' ','').split(',')
            for push_to in push_list:
                push_conf = dict_from_config_items(global_config.items(push_to))
                push_type = push_conf['push_type']
                if push_type == 'email':
                    agent.add_push_service(EmailPush(push_conf))

    if agent is None:
        agent = SimulatedAgent(market, None)

    fund.set_agent( agent, get_order_cost(global_config) )

    trader.add_fund(fund)

    market.set_verbose( verbose )
    trader.set_verbose( verbose )

    trader.run_fund(date_start, date_end, tick_period = tick_period)

    end_tick = dt.datetime.now()
    print(fund_id, 'time used:', (end_tick - start_tick))

    return trader.get_report()

def cli_run_fund_multiprocessing(config_file, start, end, options):
    verbose = '-d' in options

    # read config file
    print( 'reading config from from:', config_file)
    if not os.path.isfile(config_file):
        print('Error:', config_file, 'not eixsts\n')
        exit()

    global_config = configparser.ConfigParser()
    global_config.read(config_file, encoding='utf-8')

    main_conf = dict_from_config_items(global_config.items('main'), verbose= verbose)
    tick_period = seconds_from_str( main_conf['tick_period'] ) if ('tick_period' in main_conf) else 60
    compare_index = main_conf['compare_index'] if ('compare_index' in main_conf) else None

    date_start = date_from_str( start )
    date_end = date_from_str( end )

    start_tick = dt.datetime.now()
    fund_id_list = [fund_id for k, fund_id in global_config.items('fund_list')]
    fund_args = [[global_config, fund_id, date_start, date_end, tick_period, verbose] for fund_id in fund_id_list]
    with mp.Pool( len(fund_id_list) ) as p:
        report_list = p.map(cli_one_trader_run_one_fund, fund_args)
    end_tick = dt.datetime.now()
    print('Total time used:', (end_tick - start_tick))

    all_report = []
    for report in report_list:
        all_report += report

    out_file = None
    for option in options:
        if option.startswith('-out=') and option.endswith('.png'):
            out_file = option.replace('-out=', '')

    trader = Trader(Market(date_start, date_end))
    trader.print_report(all_report)
    trader.plot(report = all_report, compare_index= compare_index, out_file= out_file)

    print('Done.\n')

def cli_trade_backtrade(params, options):
    config_file = params[0]
    start = params[1] if len(params) > 1 else '3 years ago'
    end = params[2] if len(params) > 2 else 'yesterday'
    if '-q' in options:
        start = '3 months ago'
        end = '1 week ago'
    if '-m' in options:
        cli_run_fund_multiprocessing(config_file, start, end, options)
    else:
        cli_run_fund(config_file, start, end, options)

def cli_trade_run(params, options):
    config_file = params[0]
    cli_run_fund(config_file, 'today', 'future', options)

def cli_trade(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_trade_help()
        return

    action = params[0]
    params = params[1:]

    fund_tools = {
        'create': cli_trade_create,
        'backtrade': cli_trade_backtrade,
        'run': cli_trade_run,
    }
    if action in fund_tools.keys():
        if (len(params) == 0) or (not (params[0].endswith('.conf'))):
            print('\nError: A config file ending with .conf is expected.\n')
            return
        func = fund_tools[ action ]
        func(params, options)
    else:
        print('\nError: unknown action:', action)
        cli_trade_help()



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import pandas as pd

from ..utils import get_file_modify_time, str_now
from .agent_simulated import SimulatedAgent

class HumanAgent(SimulatedAgent):
    market = None
    conf = None

    push_service = None

    def set_verbose(self, verbose = True):
        super().set_verbose(verbose)
        if self.push_service is not None:
            self.push_service.set_verbose(verbose)

    def __init__(self, market, agent_conf = None):
        super().__init__(market, agent_conf)

        self.agent_type = 'human'

    def before_day(self):
        super().before_day()
        self.load_portoflio_from_file()



# -*- coding: utf-8; py-indent-offset:4 -*-

class MasterPush:
    services = []

    def __init__(self):
        self.services = []

    def add_service(self, push_service):
        self.services.append(push_service)

    def set_verbose(self, verbose = True):
        for k in self.services:
            k.set_verbose(verbose)

    def add_msg(self, msg):
        for k in self.services:
            k.add_msg(msg)

    def add_order(self, symbol, name, count, price, earn_str = '', comment = ''):
        for k in self.services:
            k.add_order(symbol, name, count, price, earn_str, comment)

    def flush(self):
        for k in self.services:
            k.flush()

    def clear(self):
        for k in self.services:
            k.clear()




# -*- coding: utf-8; py-indent-offset:4 -*-

import math

def earn_to_annual(earn, days):
    if days > 365:
        year = days / 365
        return math.pow(earn+1, 1/year) -1
    else:
        return earn



# -*- coding: utf-8; py-indent-offset:4 -*-

from .cli_init import cli_init
from .cli_stockpool import cli_stockpool
from .cli_finance import cli_finance
from .cli_pepb import cli_pepb
from .cli_stock import cli_stock
from .cli_index import cli_index
from .cli_indicator import cli_indicator
from .cli_strategy import cli_strategy
from .cli_trade import cli_trade
from .cli_main import cli_main, cli_main_params_options



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
import datetime as dt
import pandas as pd
import matplotlib.pyplot as plt
import talib as tl
from talib import abstract
from tabulate import tabulate

from ..core import symbol_to_name
from ..indicator import SIGN

_talib_pattern_length = {
'CDL2CROWS': 3,
'CDL3BLACKCROWS': 3,
'CDL3INSIDE': 3,
'CDL3LINESTRIKE': 4,
'CDL3OUTSIDE': 3,
'CDL3STARSINSOUTH': 3,
'CDL3WHITESOLDIERS': 3,
'CDLABANDONEDBABY': 3,
'CDLADVANCEBLOCK': 3,
'CDLBELTHOLD': 2,
'CDLBREAKAWAY': 5,
'CDLCLOSINGMARUBOZU': 1,
'CDLCONCEALBABYSWALL': 4,
'CDLCOUNTERATTACK': 2,
'CDLDARKCLOUDCOVER': 2,
'CDLDOJI': 1,
'CDLDOJISTAR': 1,
'CDLDRAGONFLYDOJI': 1,
'CDLENGULFING': 2,
'CDLEVENINGDOJISTAR': 3,
'CDLEVENINGSTAR': 3,
'CDLGAPSIDESIDEWHITE': 2,
'CDLGRAVESTONEDOJI': 1,
'CDLHAMMER': 1,
'CDLHANGINGMAN': 1,
'CDLHARAMI': 2,
'CDLHARAMICROSS': 2,
'CDLHIGHWAVE': 3,
'CDLHIKKAKE': 3,
'CDLHIKKAKEMOD': 3,
'CDLHOMINGPIGEON': 2,
'CDLIDENTICAL3CROWS': 3,
'CDLINNECK': 2,
'CDLINVERTEDHAMMER': 1,
'CDLKICKING': 2,
'CDLKICKINGBYLENGTH': 2,
'CDLLADDERBOTTOM': 5,
'CDLLONGLEGGEDDOJI': 1,
'CDLLONGLINE': 1,
'CDLMARUBOZU': 1,
'CDLMATCHINGLOW': 2,
'CDLMATHOLD': 5,
'CDLMORNINGDOJISTAR': 3,
'CDLMORNINGSTAR': 3,
'CDLONNECK': 2,
'CDLPIERCING': 2,
'CDLRICKSHAWMAN': 1,
'CDLRISEFALL3METHODS': 5,
'CDLSEPARATINGLINES': 2,
'CDLSHOOTINGSTAR': 1,
'CDLSHORTLINE': 1,
'CDLSPINNINGTOP': 1,
'CDLSTALLEDPATTERN': 3,
'CDLSTICKSANDWICH': 3,
'CDLTAKURI': 1,
'CDLTASUKIGAP': 3,
'CDLTHRUSTING': 2,
'CDLTRISTAR': 3,
'CDLUNIQUE3RIVER': 3,
'CDLUPSIDEGAP2CROWS': 3,
'CDLXSIDEGAP3METHODS': 5,
}

def list_existing_daily_files():
    file_list = []
    files = os.listdir('cache/market')
    for file in files:
        if file.endswith('_1d.csv'):
            file_list.append('cache/market/'+file)
    return file_list

def cli_pattern_list():
    all_patterns = tl.get_function_groups()['Pattern Recognition']
    table = []
    for pattern in all_patterns:
        info = _talib_pattern_length[ pattern ]
        func = abstract.Function(pattern)
        table.append([pattern, func.info['display_name'], info])
    df = pd.DataFrame(table, columns=['pattern', 'name', 'length'])
    print( tabulate(df, headers='keys', tablefmt='psql') )

def cli_pattern_stat():
    all_patterns = tl.get_function_groups()['Pattern Recognition']

    # init the counters
    pattern_counters = {}
    for pattern in all_patterns:
        pattern_counters[ pattern ] = [0, 0]

    # find all daily files
    files = list_existing_daily_files()
    n = len(files)
    i = 0
    for file in files:
        i += 1
        print('\r ... {}/{} ... searching patterns in {} ...'.format(i, n, file), end= '', flush= True)
        df = pd.read_csv(file, parse_dates=['date']).set_index('date', drop=True)
        df = df.astype(float)

        # prepare for validity exam
        diff = df['close'].diff().shift(-1).fillna(0)

        for pattern in all_patterns:
            # find the pattern signals
            func = abstract.Function(pattern)
            signal = func(df)

            # examine
            verify = SIGN( signal * diff )
            valid_signals = verify[ verify > 0 ].sum()
            invalid_signals = verify[ verify < 0 ].abs().sum()

            pattern_counters[ pattern ][0] += valid_signals
            pattern_counters[ pattern ][1] += invalid_signals
    print('\n')

    # stat result as dataframe
    table = []
    for pattern, counters in pattern_counters.items():
        sum_counter = counters[0] + counters[1]
        valid_rate = ' - ' if (sum_counter == 0) else (str(round(counters[0] / sum_counter * 100.0)).rjust(3) + " %")
        row = [pattern, counters[0], counters[1], valid_rate]
        table.append(row)
    stat_df = pd.DataFrame(table, columns=['pattern', 'valid', 'invalid', 'valid_rate'])

    stat_df.sort_values('valid_rate', ascending=False, inplace=True)

    print( tabulate(stat_df, headers='keys', tablefmt='psql') )

def find_pattern(func):
    files = list_existing_daily_files()
    for file in files:
        df = pd.read_csv(file, parse_dates=['date']).set_index('date', drop=True)
        df = df.astype(float)

        signal = func(df)
        signal = signal[ signal != 0 ]
        if len(signal) > 0:
            symbol = os.path.basename(file).replace('_daily.csv','')
            return symbol, signal, df

    return None, [], None

def get_k_value(price, pattern_date, length):
    start = pattern_date - dt.timedelta(days=20)
    end = pattern_date + dt.timedelta(days=20)
    price = price[start:end]

    dates = price.index.tolist()
    i = dates.index(pattern_date)
    pattern_dates = dates[i+1-length:i+1]

    y_list = {'y1': [], 'y2': [], 'y3': [], 'y4': [], 'c': [], 'al': []}
    
    for i in range(price.shape[0]):
        # calc bar transparency
        if price.index[i] in pattern_dates:
            y_list['al'].append(1)
        else:
            y_list['al'].append(0.3)
        
        # calc length of candle block
        if price.open[i] > price.close[i]:
            y_list['y1'].append(price.close[i])
            y_list['y2'].append(price.open[i] - price.close[i])
            y_list['c'].append('g')
        else:
            y_list['y1'].append(price.open[i])
            y_list['y2'].append(price.close[i] - price.open[i])
            y_list['c'].append('r')
        
        # calc length of candle line
        if price.high[i] > price.low[i]:
            y_list['y3'].append(price.low[i])
            y_list['y4'].append(price.high[i] - price.low[i])
        else:
            y_list['y3'].append(price.high[i])
            y_list['y4'].append(price.low[i] - price.high[i])
            
    for name, value in y_list.items():
        price.loc[:, name] = value

    return price

def draw_candle(data, title='None', out_file = None):
    # merge axis data
    x = [str(date.date()) for date in data.index]
    y1 = data['y1']
    y2 = data['y2']

    y3 = data['y3']
    y4 = data['y4']
    
    fig = plt.figure(figsize=(8, 6))

    # draw candle line with color and transparency
    for i in range(len(data)):
        plt.bar(x[i], y3[i], align='center', alpha=0)
        plt.bar(x[i], y4[i], align='center', color=data['c'][i], 
                alpha=data['al'][i], bottom=y3[i], width=0.1)

    # draw candle block with color and transparency
    for j in range(len(data)):
        plt.bar(x[j], y1[j], align='center', alpha=0)
        plt.bar(x[j], y2[j], align='center', color=data['c'][j], 
                alpha=data['al'][j], bottom=y1[j])
    
    _min = data[['open', 'high', 'low', 'close']].min().min() - 1.0
    _max = data[['open', 'high', 'low', 'close']].max().max() + 1.0

    plt.ylim(_min, _max)
    plt.margins(x=0.0, y=0.05)
    plt.title(title)
    plt.xticks(rotation=45)

    if out_file is None:
        plt.show()
    else:
        plt.savefig(out_file)

# exam pattern of a day, return pattern name
def discern_pattern(daily_df, date):
    price = daily_df[ daily_df.index <= date ]
    pattern_function = tl.get_function_groups()['Pattern Recognition']
    for name in pattern_function:
        func = abstract.Function(name)
        signal_series = func(price)
        if signal_series[-1] != 0:
            yield func.info['display_name']

def pattern_plot(pattern, out_file = None):
    all_patterns = tl.get_function_groups()['Pattern Recognition']
    pattern = all_patterns[int(pattern) % len(all_patterns)] if pattern.isdigit() else pattern
    if pattern not in all_patterns:
        print('\nError: pattern "{}" not exists'.format(pattern))
        return False

    func = abstract.Function(pattern)
    symbol, signal, df = find_pattern(func)

    if len(signal) > 0:
        symbol = symbol.replace('_1d.csv', '')
        name = symbol_to_name(symbol)

        # plot the pattern
        length = _talib_pattern_length[pattern]
        k_value = get_k_value(df, signal.index[0], length)
        title = '{} - {} : {}'.format(symbol, name, func.info['display_name'])
        draw_candle(k_value, title, out_file= out_file)
        return True

    else:
        print('\nPattern "{}" not found after searching all daily data in cache/market .'.format(pattern))
        print('You can try again after download more daily data.\n')
        return False

def cli_pattern_help():
    syntax_tips = '''Syntax:
    __argv0__ pattern list
    __argv0__ pattern stat
    __argv0__ pattern demo <pattern>

Actions:
    list ........................... list pattern supported in this tool
    stat ........................... count all the patterns and effectiveness
    demo ........................... plot pattern as demo

Pattern:
    i .............................. index of patterns, range 0 - 60
    name ........................... name of pattern, like CDL2CROWS

Example:
    __argv0__ pattern list
    __argv0__ pattern demo 1
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print( syntax_tips )

def cli_pattern_plot(params, options):
    out_file = None
    for option in options:
        if option.startswith('-out=') and option.endswith('.png'):
            out_file = option.replace('-out=', '')

    if len(params) > 0:
        if params[0] == 'all':
            all_patterns = tl.get_function_groups()['Pattern Recognition']
            for pattern in all_patterns:
                out_file = 'output/{}.png'.format(pattern)
                if pattern_plot(pattern, out_file= out_file):
                    print('pattern "{}" saved to: {}'.format(pattern, out_file))
        else:
            pattern = params[0]
            if pattern_plot(pattern, out_file= out_file) and (out_file is not None):
                print('pattern "{}" saved to: {}'.format(pattern, out_file))
    else:
        print('\nError: pattern index or name exptected.')

def cli_pattern(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_pattern_help()
        return

    action = params[0]
    params = params[1:]

    if action == 'list':
        cli_pattern_list()

    elif action == 'stat':
        cli_pattern_stat()

    elif action in ['demo', 'plot']:
        cli_pattern_plot(params, options)

    else:
        print('\nError: invalid action:', action)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os

from ..utils import datetime_today
from .lang import LANG

class Strategy:
    fund = None
    name = ''
    conf = {}

    max_stocks = 10
    max_weight = 1.0
    stop_loss = 0.0
    stop_earn = 10000.0

    def __init__(self, fund, strategy_file):
        fund.add_strategy(self)

        self.fund = fund
        self.name = os.path.basename(strategy_file)
        self.conf = fund.conf

        self.max_stocks = 10
        self.max_weight = 1.0
        self.stop_loss = 1 + (-1.0)
        self.stop_earn = 1 + (10000.0)

        if 'max_stocks' in self.conf:
            self.max_stocks = int(self.conf['max_stocks'])
        if 'max_weight' in self.conf:
            self.max_weight = float(self.conf['max_weight'])
        if 'stop_loss' in self.conf:
            self.stop_loss = 1 + float(self.conf['stop_loss'])
        if 'stop_earn' in self.conf:
            self.stop_earn = 1 + float(self.conf['stop_earn'])

class BasicStrategy( Strategy ):
    symbol_signal = {}

    def __init__(self, fund, strategy_file):
        super().__init__(fund, strategy_file)

        self.targets = self.select_stock()
        self.symbol_signal = {}

        market = fund.market
        agent = fund.agent
        portfolio = agent.get_portfolio()

        concerned_stocks = list(set(self.targets) | set(portfolio.positions.keys()))
        concerned_stocks.sort()
        market.watch( concerned_stocks )

        self.schedule_task(fund.trader)

    def schedule_task(self, trader):
        trader.run_daily(self.trade, None, time='14:30')
        trader.run_on_bar_update(self.trade, None)

    def select_stock(self):
        pass

    def gen_trade_signal(self, symbol, init_data = False):
        pass

    def get_signal_comment(self, symbol, signal):
        pass

    def init_trade_signal(self, symbol):
        return symbol, self.gen_trade_signal(symbol, True)

    def get_trade_decision(self, symbol, market, portfolio, max_pos_per_stock):
        if not market.allow_trading(symbol, market.current_date):
            return symbol, 0, ''

        price = market.get_price(symbol, market.current_date)
        if price < 0.01:
            return symbol, 0, ''

        if market.current_date < datetime_today():
            date_signal = self.symbol_signal[ symbol ]
            if market.current_date in date_signal:
                trade_signal = date_signal.loc[ market.current_date ]
            else:
                trade_signal = self.gen_trade_signal(symbol, init_data=False).iloc[-1]
        else:
            trade_signal = self.gen_trade_signal(symbol, init_data=False).iloc[-1]

        if (symbol in portfolio.positions):
            cost = portfolio.positions[ symbol ].cost
            if (trade_signal < 0):
                return symbol, -1, LANG('signal') + ': ' + self.get_signal_comment(symbol, trade_signal)
            elif (trade_signal > 0):
                # will handle it below
                pass
            else:
                # only stop loss/earn when no signal to buy or sell
                # or else, may need buy again after sell
                if price <= cost * self.stop_loss:
                    stop_percent = round(100*(self.stop_loss-1),2)
                    return symbol, -1, LANG('stop loss') + ': {} %'.format(stop_percent)
                elif price >= cost * self.stop_earn:
                    stop_percent = round(100*(self.stop_earn-1),2)
                    return symbol, -1, LANG('stop earn') + ' {} %'.format(stop_percent)

        if (trade_signal > 0) and (symbol in self.targets):
            stock_value = 0.0
            if symbol in portfolio.positions:
                stock_value = price * portfolio.positions[ symbol ].shares
            if stock_value < max_pos_per_stock:
                return symbol, 1, LANG('signal') + ': ' + self.get_signal_comment(symbol, trade_signal)

        return symbol, 0, ''

    def trade(self, param = None):
        fund = self.fund
        market = fund.market
        agent = fund.agent
        portfolio = agent.get_portfolio()

        total_value = portfolio.total_value()
        max_stocks = min(self.max_stocks, len(self.targets))
        max_pos_per_stock = total_value / max_stocks * self.max_weight

        # we merge the targets and in stock shares, but keep tarets ahead of others with higher priority to buy
        concerned_stocks = self.targets + list(set(portfolio.positions.keys()) - set(self.targets))

        # make sure the daily data in cache of market object
        market.watch( concerned_stocks )

        # for those symbol witout signal, we need init the data
        symbol_list = list(set(concerned_stocks) - self.symbol_signal.keys())
        symbol_signal_list = [self.init_trade_signal(symbol) for symbol in symbol_list]
        for symbol, signal in symbol_signal_list:
            self.symbol_signal[ symbol ] = signal

        # find out the stocks to sell or buy
        decision_list = [self.get_trade_decision(symbol, market, portfolio, max_pos_per_stock) for symbol in concerned_stocks]

        # now sell first, we may need the cash to buy other stocks
        for symbol, signal, reason in decision_list:
            if signal < 0:
                agent.order_target(symbol, 0, comment = reason)

        # now buy
        for symbol, signal, reason in decision_list:
            if portfolio.count() >= max_stocks:
                break
            if (signal > 0) and (symbol not in portfolio.positions):
                agent.order_target_value(symbol, max_pos_per_stock, comment = reason)



# -*- coding: utf-8 -*-

import pandas as pd

from .basic import CROSS, REF
from .ma import EMA
from ..core.indicator_signal import register_signal_indicator

# Money Flow Index, by Gene Quong & Avrum Soudack
# --------------------------------------------
def MFI(close, high, low, volume, length = 14, n = 6):
    tp = (high + low + close) / 3.0
    mf = tp * volume
    diff = tp - REF(tp, 1)

    tdf = pd.DataFrame({"diff": 0, "rmf": mf, "+mf": 0, "-mf": 0})
    tdf.loc[(diff > 0), "diff"] = 1
    tdf.loc[tdf["diff"] == 1, "+mf"] = mf

    tdf.loc[(diff < 0), "diff"] = -1
    tdf.loc[tdf["diff"] == -1, "-mf"] = mf

    psum = tdf["+mf"].rolling(length, min_periods=1).sum()
    nsum = tdf["-mf"].rolling(length, min_periods=1).sum()
    mfi = 100 * psum / (psum + nsum)

    return mfi

def signal_mfi(df, inplace=False, length=14, n=6):
    mfi = MFI(df.close, df.high, df.low, df.volume, length=length, n=n)
    if inplace:
        df['mfi'], df['mfi_upper'], df['mfi_lower'] = mfi, 80, 20
    buy_signal = CROSS(mfi, 80) < 0
    sell_signal = CROSS(mfi, 20) > 0
    return buy_signal.astype(int) - sell_signal.astype(int)

register_signal_indicator('mfi', signal_mfi, ['mfi','mfi_upper','mfi_lower'], 'MFI', 'volume')



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
import tabulate as tb

from ..utils import symbol_normalize, date_limit_from_options
from ..data_source import download_cn_index_stocks_list
from ..core import symbol_to_name
from ..core import get_order_cost
from ..core import list_signal_indicators
from ..core import get_cn_index_list_df, get_hk_index_list_df, get_us_index_list_df, get_world_index_list_df, get_index_daily
from ..core import Stock

def cli_index_help():
    syntax_tips = '''Syntax:
    __argv0__ index list <region>
    __argv0__ index plot <symbol> [<from_date>] [<to_date>] [<options>]
    __argv0__ index stocks <symbol> [-out=<out.csv>]

<action>
    list ....................... list index, following param: world, cn, us, hk, etc.
    plot ....................... plot the index
    stocks ..................... list stocks of this index

<symbol> ....................... stock symbol, like 600036
<from_date> .................... default from 3 year ago
<to_date> ...................... default to yesterday

<options> ...................... options to plot the stock
    -ma, -macd, -kdj, etc. ..... add some indicators
    -all ....................... add all indicators
    -mix ....................... show trade result if mix trade signal of the indicators
    -top5 ...................... show only top 5 indicators according to result

Example:
    __argv0__ index list world
    __argv0__ index list cn
    __argv0__ index list us

    __argv0__ index plot sh000300 -ma
    __argv0__ index plot sh000300 20180101 -ma -macd -kdj
    __argv0__ index plot sh000300 20180101 20210101 -ma -macd -kdj

    __argv0__ index stocks sh000300 -out=stockpool/sh000300_stocks.csv

'''.replace('__argv0__',os.path.basename(sys.argv[0]))
    print(syntax_tips)

def cli_index_list(params, options):
    list_funcs = {
        'world': get_world_index_list_df,
        'cn': get_cn_index_list_df,
        'us': get_us_index_list_df,
        'hk': get_hk_index_list_df,
    }
    if (len(params) == 0) or (not (params[0] in list_funcs)):
        print('Error: index list expected param:', ', '.join(list(list_funcs.keys())))
        cli_index_help()
        return
    func = list_funcs[ params[0] ]
    df = func()
    print( tb.tabulate(df, headers='keys', showindex=True, tablefmt='psql') )
    print( 'Totally', df.shape[0], 'records.\n')

def cli_index_plot(params, options):
    symbol = symbol_normalize(params[0])
    name = symbol_to_name(symbol)

    date_start, date_end, limit = date_limit_from_options(options)

    df = get_index_daily(symbol)
    df = df[ df.index >= date_start ]
    df = df[ df.index < date_end ]

    stock = Stock(symbol, name, df)

    all_indicators = list_signal_indicators()

    indicators = []
    topN = 5
    inplace = True
    out_file = None
    for option in options:
        if option.startswith('-top'):
            topN = int(option.replace('-top',''))
        elif option.startswith('-out=') and (option.endswith('.png') or option.endswith('.jpg')):
            out_file = option.replace('-out=', '')
        else:
            k = option.replace('-','')
            if (k in all_indicators) and (not k in indicators):
                indicators.append(k)
    if '-all' in options:
        indicators = all_indicators
        inplace = False

    mix = '-mix' in options

    # add indicators and calculate performance
    stock.add_indicator(indicators, mix=mix, inplace= inplace, order_cost = get_order_cost())

    rank_df = stock.rank_indicator(by = 'final')
    if rank_df.shape[0] > 0:
        print(rank_df)

    # find top indicators
    ranked_indicators = rank_df['indicator'].tolist()
    other_indicators = ranked_indicators[topN:]
    df.drop(columns = other_indicators, inplace=True)

    stock.plot(out_file= out_file)

def cli_index_stocks(params, options):
    if len(params) > 0:
        symbol = params[0]
        if symbol.startswith('sh') or symbol.startswith('sz'):
            df = download_cn_index_stocks_list(symbol)
            print( tb.tabulate(df, headers='keys', showindex=True, tablefmt='psql') )
            print( 'Totally', df.shape[0], 'records.\n')

            for option in options:
                if ('-out=' in option) and option.endswith('.csv'):
                    out_csv_file = option.replace('-out=','')
                    df.to_csv(out_csv_file, index=False)
                    print('Saved to:', os.path.abspath(out_csv_file))
                    return
            return

    print('Expect param: China index symbol, with prefix sh or sz, like sh000300, sz399991, etc.\n')

def cli_index(params, options):

    if (len(params) == 0) or (params[0] == 'help'):
        cli_index_help()
        return

    action = params[0]
    params = params[1:]

    if action == 'list':
        cli_index_list(params, options)

    elif action in ['plot', 'view', 'show']:
        cli_index_plot(params, options)

    elif action == 'stocks':
        cli_index_stocks(params, options)

    else:
        print('invalid action:', action)
        cli_index_help()



# -*- coding: utf-8; py-indent-offset:4 -*-

import datetime as dt
from .lang import LANG

class PushBase:
    conf = None
    msg_queue = []
    verbose = False

    def __init__(self, conf):
        self.conf = conf.copy()

    def set_verbose(self, verbose = True):
        self.verbose = verbose

    def add_msg(self, msg):
        self.msg_queue.append(msg)
        pass

    def add_order(self, symbol, name, count, price, earn_str = '', comment = ''):
        trade = LANG('You can buy') if count > 0 else LANG('You can sell')
        msg = '{} {}, {} {} * {}, {} {}'.format(symbol, name, trade, count, price, earn_str, comment)
        self.msg_queue.append(msg)

    def clear(self):
        self.msg_queue = []

    def flush(self):
        if not self.msg_queue:
            return

        line = '-' * 40
        content = '\n\n'.join(self.msg_queue)
        content += '\n' + line + '\n' + dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S') + '\n'

        if self.verbose:
            print(content)

        self.msg_queue = []



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys

import pandas as pd
import tabulate as tb

from ..core.data_cache import get_pepb_symmary_df
from ..utils import filter_with_options, sort_with_options

def cli_pepb_help():
    syntax_tips = '''Syntax:
    __argv0__ pepb <action> <symbols | symbos.csv | all> [options]

Action:
    update ................................. download PE/PB data and update PE/PE calculation
    view ................................... view PE/PB data with filters

Symbols:
    <symbols> .............................. stock symbol list, like 600036 000002
    <symbols.csv> .......................... stock symbol list csv file
    all .................................... all symbols

Options:
    -sortby=<col> .......................... sort by the column

    -pe=<from>-<to> ........................ sort by PE
    -pb=<from>-<to> ........................ sort by PB
    -pe_pos=<from>-<to> .................... sort by PE % position
    -pb_pos=<from>-<to> .................... sort by PB % position

    -out=<out.csv> ......................... export selected stocks into <out.csv> file

Example:
    __argv0__ pepb view 600036 000002 600276
    __argv0__ pepb view my-stocks.csv -pe_pos
    __argv0__ pepb view all
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print( syntax_tips )

def cli_pepb_view(params, options, force_update = False):
    if params[0] == 'all':
        df = get_pepb_symmary_df()
    else:
        if params[0].endswith('.csv'):
            stock_df = pd.read_csv(params[0], dtype=str)
            symbols = stock_df['symbol'].tolist()
        else:
            symbols = params
        df = get_pepb_symmary_df(symbols)

    total_n = df.shape[0]

    # now filter
    df = filter_with_options(df, options)
    filtered_n = df.shape[0]

    # now sort
    df = sort_with_options(df, options, by_default='pb_pos')

    print( tb.tabulate(df, headers='keys', tablefmt='psql') )

    print('{} out of {} records selected.'.format(filtered_n, total_n))

    out_xlsx_file = ''
    out_csv_file = ''
    for k in options:
        if k.startswith('-out='):
            if k.endswith('.xlsx'):
                out_xlsx_file = k.replace('-out=', '')
            if k.endswith('.csv'):
                out_csv_file = k.replace('-out=', '')

    if out_csv_file:
        df = df[['symbol', 'name']]
        df.to_csv(out_csv_file, index= False)
        print('Exported to:', out_csv_file)
        print(df)

    if out_xlsx_file:
        name_col = df.pop('name')
        df.insert(1, 'name', name_col)
        df = df.rename(columns={
            'symbol':'代码',
            'name': '名称',
            'pe': '市盈率',
            'pb': '市净率',
            'pe_pos': '市盈率百分位',
            'pb_pos': '市净率百分位',
            'pe_ratio': '市盈率均值比率',
            'pb_ratio': '市净率均值比率',
        })
        df.to_excel(out_xlsx_file)
        print('Exported to:', out_xlsx_file)

    print('')

def cli_pepb(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_pepb_help()
        return

    action = params[0]
    params = params[1:]

    if action == 'update':
        cli_pepb_view(params, options, force_update= True)

    elif action in ['view', 'show', 'filter']:
        cli_pepb_view(params, options, force_update= False)

    else:
        print('invalid action:', action)
        cli_pepb_help()



# -*- coding: utf-8 -*-

import pandas as pd

from ..core.indicator_signal import register_signal_indicator, signal_no_dup

# Main Fund Flow Index, by Liming Xie
def signal_mffi(df, inplace=False, days= 1, percent= 5.0):
    mffi = df['main_pct'] if ('main_pct' in df.columns) else pd.Series(0, index=df.index)
    if inplace:
        df['mffi'] = mffi

    # main fundflow in/out for over two days
    signal = pd.Series(0, index=mffi.index)
    n = len(signal.tolist())
    for i in range(days-1, n):
        for j in range(0, days):
            if mffi.iloc[i-j] > percent:
                signal.iloc[i] += 1
            elif mffi.iloc[i-j] < -percent:
                signal.iloc[i] += -1
        signal.iloc[i] //= days

    return signal #signal_no_dup(signal)

register_signal_indicator('mffi', signal_mffi, ['mffi'], 'MFFI', 'volume')



# -*- coding: utf-8; py-indent-offset:4 -*-

import os

import pandas as pd

from ..utils import datetime_today, get_file_modify_time, str_now
from .portfolio import Portfolio
from .order_cost import OrderCost
from .push_master import MasterPush
from .stock import Stock
from .lang import LANG

class SimulatedAgent:
    conf = {}
    agent_type = ''

    market = None
    order_cost = None
    push_service = None

    last_order_time = None
    order_file = ''
    portfolio_save_file = ''
    portfolio_load_file = ''
    last_load_modified_time = None

    portfolio = None
    sell_count = 0
    buy_count = 0

    verbose = False

    def set_verbose(self, verbose = True):
        self.verbose = verbose

    def set_market(self, market):
        self.market = market

    def add_push_service(self, push_servcie):
        self.push_service.add_service(push_servcie)

    def set_order_cost(self, order_cost):
        self.order_cost = order_cost

    def __init__(self, market, agent_conf = {}):
        self.agent_type = 'simulated'
        self.market = market
        self.conf = agent_conf

        self.portfolio = Portfolio(market)
        self.order_cost = OrderCost(0.001, 0.0003, 0.0003, 5.0)
        self.push_service = MasterPush()

        self.sell_count = 0
        self.buy_count = 0

        if 'order' in self.conf:
            self.order_file = self.conf['order']
        if 'portfolio_save' in self.conf:
            self.portfolio_save_file = self.conf['portfolio_save']
        if 'portfolio_load' in self.conf:
            self.portfolio_load_file = self.conf['portfolio_load']

    def load_portoflio_from_file(self):
        if os.path.isfile(self.portfolio_load_file):
            load_modified_time = get_file_modify_time(self.portfolio_load_file)
            stock_df = pd.read_csv(self.portfolio_load_file, dtype= str)
            stock_df = stock_df.astype({
                'shares': 'float64',
                'cost': 'float64',
            })
            self.portfolio.from_dataframe( stock_df )
            self.last_load_modified_time = load_modified_time
            print('\n... {} |'.format(str_now()), 'loaded:', self.portfolio_load_file, load_modified_time)
            if self.verbose:
                stock_df['price'] = self.market.get_real_price(stock_df['symbol'].tolist())
                stock_df['value'] = stock_df['price'] * stock_df['shares']
                print(stock_df)

    def get_portfolio(self):
        return self.portfolio

    def get_transaction_count(self):
        return self.buy_count, self.sell_count

    def init_portfolio(self, start_cash):
        self.portfolio = Portfolio(self.market)
        self.portfolio.available_cash = start_cash
        self.last_order_time = self.market.current_time
        if not (not self.portfolio_load_file):
            self.load_portoflio_from_file()
            self.portfolio.print()

    def transfer_cash(self, amount, bank_to_security = True):
        if bank_to_security:
            self.portfolio.available_cash += amount
        else:
            self.portfolio.available_cash -= amount
        if self.portfolio.available_cash < 0:
            self.portfolio.available_cash = 0.0

    def before_day(self):
        self.buy_count = 0
        self.sell_count = 0
        for symbol, stock in self.portfolio.positions.items():
            stock.sellable = True

    def after_day(self):
        pass

    def before_tick(self):
        pass

    def after_tick(self):
        if self.market.current_date < datetime_today():
            self.push_service.clear()
        else:
            self.push_service.flush()

            if self.portfolio_save_file:
                if os.path.isfile(self.portfolio_save_file):
                    save_modified_time = get_file_modify_time(self.portfolio_save_file)
                    need_save = (self.last_order_time > save_modified_time)
                else:
                    need_save = True

                if need_save:
                    stock_df = self.portfolio.to_dataframe()
                    stock_df.to_csv( self.portfolio_save_file, index= False)
                    print('\n... {} |'.format(str_now()), 'updated:', self.portfolio_save_file)

                    if self.portfolio_save_file == self.portfolio_load_file:
                        self.last_load_modified_time = get_file_modify_time(self.portfolio_load_file)

            if os.path.isfile(self.portfolio_load_file):
                load_modified_time = get_file_modify_time(self.portfolio_load_file)
                need_load = (load_modified_time > self.last_load_modified_time)
                if need_load:
                    self.load_portoflio_from_file()

    def exec_order(self, symbol, real_count, real_price, earn_str = '', comment = ''):
        market = self.market
        name = market.get_name(symbol)

        self.last_order_time = market.current_time

        str_now = market.current_time.strftime('%Y-%m-%d %H:%M:%S')
        trade = LANG('buy') if real_count > 0 else LANG('sell')
        msg = '{} | {} {} | {}: {} * {} | {} {}'.format(str_now, symbol, name, trade, round(real_count), round(real_price,2), earn_str, comment)
        if self.verbose:
            print('\r'+msg)

        if self.order_file:
            fp = open(self.order_file, 'a+')
            fp.write(msg + '\n')
            fp.close()

        if self.push_service is not None:
            market = self.market
            name = market.get_name(symbol)
            self.push_service.add_order(symbol, name, real_count, real_price, earn_str, comment)

    def order(self, symbol, count, price = None, comment = ''):
        market = self.market
        if not price:
            price = market.get_price(symbol)
            if price <= 0:
                return

        p = self.portfolio

        # make sure the count is within our budget
        count = min(count, p.available_cash / price)

        real_price = market.get_real_price(symbol)
        real_count = price * count / real_price

        cost = 0
        if symbol in p.positions:
            stock = p.positions[symbol]
            target_count = stock.shares + count
            if target_count > 0:
                # buy, or not sell all, a hand must be times of 100
                real_count = int(real_count / 100) * 100
                if real_count == 0:
                    if self.verbose:
                        print('must buy/sell no less than 1 hand', symbol, count, real_count, real_price, p.available_cash)
                    return

                count = real_count * real_price / price

                stock.cost = (stock.cost * stock.shares + price * count) / target_count
                stock.shares += count
            else:
                if stock.sellable:
                    # sell all
                    count = - stock.shares
                    real_count = price * count / real_price

                    cost = stock.cost
                    stock.cost = 0.0
                    stock.shares = 0
                    p.history[symbol] = stock
                    del p.positions[symbol]
                else:
                    if self.verbose:
                        print('stock just buy in, not sellable', symbol, stock.name)
                    return

        elif count > 0:
            name = market.get_name(symbol)
            # a hand must be times of 100
            real_count = int(real_count / 100) * 100
            if real_count == 0:
                if self.verbose:
                    print('must buy no less than 1 hand', symbol, name, count, real_count, real_price, p.available_cash)
                return

            count = real_count * real_price / price

            if symbol in p.history:
                stock = p.history[symbol]
                del p.history[symbol]
            else:
                stock = Stock(symbol, name)

            stock.cost = price
            stock.shares = count
            stock.sellable = (market.sell_rule == 'T+0')

            p.positions[symbol] = stock

        else:
            raise ValueError('attempt sell shares with no position')
            return

        # calculate trade cost
        trade_cost = 0.0
        trade_value = price * max(count, -count)
        if count > 0:
            trade_cost = trade_value * (self.order_cost.open_commission)
        else:
            trade_cost = trade_value * (self.order_cost.close_commission + self.order_cost.close_tax)
        if cost < self.order_cost.min_commission:
            trade_cost = self.order_cost.min_commission

        p.available_cash = round(p.available_cash  - price * count - trade_cost, 2)

        if real_count > 0:
            self.buy_count += 1
        elif real_count < 0:
            self.sell_count += 1

        earn = round(price - cost, 2)
        earn_str = '--> {} %, '.format( round(earn / cost * 100,2) ) if cost > 0 else ''
        self.exec_order(symbol, round(real_count), round(real_price,3), earn_str, comment)

    def order_target(self, symbol, target_count, price = None, comment = ''):
        if target_count < 0:
            return
        if not price:
            market = self.market
            price = market.get_price(symbol)
        if price <= 0:
            return
        p = self.portfolio
        if symbol in p.positions:
            stock = p.positions[symbol]
            count = target_count - stock.shares
        else:
            if target_count == 0:
                return
            count = target_count
        self.order(symbol, count, price, comment)

    def order_value(self, symbol, value, price = None, comment = ''):
        if not price:
            market = self.market
            price = market.get_price(symbol)
        if price <= 0:
            return
        count = value / price
        if count > 0:
            self.order(symbol, count, price, comment)

    def order_target_value(self, symbol, target_value, price = None, comment = ''):
        if not price:
            market = self.market
            price = market.get_price(symbol)
        if price <= 0:
            return
        target_count = target_value / price
        self.order_target(symbol, target_count, price, comment)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
import datetime as dt

from ..utils import date_from_str, dict_from_config_items
from ..core import get_lang, get_hiquant_conf
from ..core import Market, Trader, Fund, SimulatedAgent

def get_strategy_template():
    return '''
# -*- coding: utf-8; py-indent-offset:4 -*-
import pandas as pd
import hiquant as hq

class MyStrategy( hq.BasicStrategy ):
    def __init__(self, fund):
        super().__init__(fund, __file__)
        self.max_stocks = 10
        self.max_weight = 1.2
        self.stop_loss = 1 + (-0.10)
        self.stop_earn = 1 + (+0.20)

    def schedule_task(self, trader):
        trader.run_daily(self.trade, None, time='09:30')
        trader.run_on_bar_update(self.trade, None)

    def select_stock(self):
        return ['AAPL', 'MSFT', 'AMZN', 'TSLA', '0700.HK']

    def gen_trade_signal(self, symbol, init_data = False):
        market = self.fund.market
        if init_data:
            df = market.get_daily(symbol)
        else:
            df = market.get_daily(symbol, end = market.current_date, count = 26+9)

        dif, dea, macd_hist = hq.MACD(df.close, fast=12, slow=26, signal=9)
        signal = pd.Series( hq.CROSS(dif, dea), index=df.index )

        # Notice!!! Important !!!
        # if we used the close price of the day to calc indicator,
        # to avoid "future data or function" in backtesting, it should not be used for today's trading
        # either we trade at 14:30 before market close
        # or, shift(1) and trade next morning
        return signal.shift(1).fillna(0)

    def get_signal_comment(self, symbol, signal):
        return 'MACD golden cross' if (signal > 0) else 'MACD dead cross'

def init(fund):
    strategy = MyStrategy(fund)

if __name__ == '__main__':
    backtest_args = dict(
        #start_cash= 1000000.00,
        #date_start= hq.date_from_str('3 years ago'),
        #date_end= hq.date_from_str('yesterday'),
        #out_file= 'output/demo.png',
        #parallel= True,
        compare_index= '^GSPC',
    )
    hq.backtest_strategy( MyStrategy, **backtest_args )

'''

def cli_strategy_create(params, options):
    if (len(params) == 0) or (not (params[0].endswith('.py'))):
        print('\nError: A strategy filename ending with .py is expected.\n')
        return

    file_to_create = params[0]
    if os.path.isfile(file_to_create):
        print('\nError: file already exists:', file_to_create)
        return

    template_content = get_strategy_template()
    fp = open(file_to_create, 'w')
    fp.write(template_content)
    fp.close()

    print('-' * 80)
    print( template_content )
    print('-' * 80)

    print( 'Strategy file created:\n  ', os.path.abspath(file_to_create))
    print( '\nPlease edit file content before running.' )

def cli_strategy_backtest(params, options):
    if (len(params) == 0) or (not (params[0].endswith('.py'))):
        print('\nError: A strategy filename ending with .py is expected.\n')
        return

    strategy_file = params[0]
    start = params[1] if len(params) > 1 else '3 years ago'
    end = params[2] if len(params) > 2 else 'yesterday'
    if '-q' in options:
        start = '3 months ago'
        end = '1 week ago'
    verbose = '-d' in options
    date_start = date_from_str( start )
    date_end = date_from_str( end )

    start_tick = dt.datetime.now()

    # market is a global singleton
    market = Market(date_start - dt.timedelta(days=90), date_end)
    trader = Trader(market)

    fund_conf = {
        'strategy': strategy_file,
    }
    fund = Fund(market, trader, fund_conf)
    agent = SimulatedAgent(market)
    fund.set_agent(agent)
    fund.set_name( os.path.basename(strategy_file) )
    fund.set_start_cash( 1000000.00 )
    trader.add_fund(fund)
    market.set_verbose( verbose )
    trader.set_verbose( verbose )
    trader.run_fund(date_start, date_end)
    trader.print_report()
    end_tick = dt.datetime.now()
    print('time used:', (end_tick - start_tick))

    config = get_hiquant_conf()
    main_conf = dict_from_config_items(config.items('main'))
    if 'compare_index' in main_conf:
        compare_index = main_conf[ 'compare_index' ]
    else:
        compare_index = 'sh000300' if (get_lang() == 'zh') else '^GSPC'

    out_file = None
    for option in options:
        if option.startswith('-out=') and option.endswith('.png'):
            out_file = option.replace('-out=', '')
        elif option.startswith('-cmp='):
            compare_index = option.replace('-cmp=', '')

    # compare with an index
    trader.plot(compare_index= compare_index, out_file= out_file)

    print('Done.\n')

def cli_strategy_help():
    syntax_tips = '''Syntax:
    __argv0__ strategy <action> <my_strategy.py> [options]

Actions:
    create ................ create a strategy python file from template
    backtest .............. backtest a strategy with default settings

<my_strategy.py> ............ a strategy python file

[options]
    -cmp=<compare index> .... compare with the index
    -out=<output.png> ....... output the plot to png file

Example:
    __argv0__ strategy create strategy/my_strategy.py
    __argv0__ strategy backtest strategy/my_strategy.py 20180101 -cmp=sh000300 -out=output/backtest.png

Alias:
    __argv0__ backtest strategy/my_strategy.py 20180101
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print( syntax_tips )

def cli_strategy(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_strategy_help()
        return

    action = params[0]
    params = params[1:]

    if action == 'create':
        cli_strategy_create(params, options)

    elif action == 'backtest':
        cli_strategy_backtest(params, options)

    else:
        print('\nError: unknown action:', action)
        cli_strategy_help()



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys

from ..core import init_hiquant_conf

def cli_init_help():
    syntax_tips = '''Syntax:
    __argv0__ init <folder>

Example:
    __argv0__ init myProj
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print(syntax_tips)

def cli_init_folder(params, options):
    folder = params[0]
    if os.path.exists(folder):
        print('Folder', folder, 'already exists.\n')
        return

    # init following folder structure
    '''
hiquant.conf
cache/
cache/finance/
cache/market/
cache/fund/
cache/pepb/
cache/bank/
cache/A_index_all.csv
cache/A_stock_all.csv
etc/
stockpool/
strategy/
data/
output/
log/
'''

    # init working folder first
    print('Initializing hiquant project folder ...')
    print('  Creating', folder)
    os.mkdir(folder)

    # init sub folders
    for sub in [
        'cache',
        'cache/finance',
        'cache/market',
        'cache/fund',
        'cache/pepb',
        'cache/bank',
        'etc',
        'stockpool',
        'strategy',
        'data',
        'output',
        'log',
        ]:
        subfolder = folder + '/' + sub
        print('  Creating', subfolder)
        os.mkdir(subfolder)

    # init hiquant config file
    conf_file = folder + '/hiquant.conf'
    print('  Creating', conf_file)
    init_hiquant_conf(conf_file)

    print('Done.\n')

def cli_init(params, options):

    if (len(params) == 0) or (params[0] == 'help'):
        cli_init_help()
        return

    cli_init_folder(params, options)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
import datetime as dt
import pandas as pd
import tabulate as tb
import matplotlib.pyplot as plt

from ..utils import date_from_str
from ..core import get_cn_stock_fund_flow_rank, get_all_symbol_name
from ..core import Market

def cli_fundflow_rank(symbols, options):
    days = '1d'
    for d in ['-1d', '-3d', '-5d', '-10d']:
        if d in options:
            days = d.replace('-', '')

    df = get_cn_stock_fund_flow_rank(days= days)
    df = df[ df['symbol'].isin(symbols) ]
    #df = df[ (df['main_pct'] > 3) | (df['main_pct'] < -3) ]

    fundflow_col = 'main_pct' if ('-pct' in options) else 'main_fund'
    df = df.sort_values(by=fundflow_col, ascending=False).reset_index(drop=True)
    df = df.head(50)

    df = df.drop(columns=['super_pct', 'large_pct', 'medium_pct', 'small_pct'])
    print( tb.tabulate(df, headers='keys', tablefmt='psql') )

def trim_axs(axs, N):
    """
    Reduce *axs* to *N* Axes. All further Axes are removed from the figure.
    """
    axs = axs.flat
    for ax in axs[N:]:
        ax.remove()
    return axs[:N]

def cli_fundflow_view(symbols, options):
    days = '1d'
    for d in ['-1d', '-3d', '-5d', '-10d']:
        if d in options:
            days = d.replace('-', '')

    df = get_cn_stock_fund_flow_rank(days= days)
    df = df[ df['symbol'].isin(symbols) ]
    df = df.head(100)

    fundflow_col = 'main_pct' if ('-pct' in options) else 'main_fund'
    df = df.sort_values(by= fundflow_col, ascending= False).reset_index(drop= True)

    symbols = df.symbol.tolist()
    all_symbol_name = get_all_symbol_name()

    plt.rcParams.update({
        'font.family': 'sans-serif',
        'font.sans-serif': ['SimHei'], # Chinese font
        'axes.unicode_minus': False,
        'axes.titlesize': 'medium', # large, medium, small
    })

    market = Market(start= date_from_str('3 month ago'), end=date_from_str('today'))
    market.watch( symbols )
    market.update_daily_realtime()

    figsize = (18, 9)
    cols = 10
    rows = len(symbols) // cols + 1
    axs = plt.figure(figsize=figsize, constrained_layout=True).subplots(rows, cols)
    axs = trim_axs(axs, len(symbols))
    for ax, symbol in zip(axs, symbols):
        ax.set_title(symbol + ' - ' + all_symbol_name[symbol])

        df = market.get_daily(symbol, count=30)
        df.index = df.index.strftime('%Y-%m%d')

        fundflow = df[ fundflow_col ]
        color = ['r' if v >= 0 else 'g' for v in fundflow]
        ax.bar(fundflow.index, fundflow, color=color)

        price = df['close']
        max_price, min_price = max(price), min(price)
        max_y, min_y = max(fundflow), min(fundflow)
        price_trend = (price - min_price) / (max_price - min_price) * (max_y - min_y) + min_y
        ax.plot(price_trend.index, price_trend, color='black', linewidth=0.6)

        ax.set_xticks([])

    plt.show()

def cli_fundflow(params, options):
    syntax_tips = '''Syntax:
    __argv0__ fundflow <action> <symbols | symbols.csv>

Actioin:
    rank ............................. rank the main fund flow
    view ............................. view & plot history fund flow of symbols

Options:
    -pct ............................. plot main fundflow percertage data
    -fund ............................ by default, plot main fundflow fund data
    -1d, -3d, -5d, -10d .............. rank by fund flow of the days

Example:
    __argv0__ fundflow view
    __argv0__ fundflow view 600036 000002 600276
    __argv0__ fundflow view stockpool/realtime_trade.csv -pct

'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    if (len(params) == 0) or (params[0] == 'help'):
        print(syntax_tips)
        return

    action = params[0]
    params = params[1:]

    if (len(params) == 0) or (params[0] == 'all'):
        symbols = list(get_all_symbol_name().keys())
    elif params[0].endswith('.csv'):
        stock_df = pd.read_csv(params[0], dtype=str)
        symbols = stock_df['symbol'].tolist()
    else:
        symbols = params

    if action == 'rank':
        cli_fundflow_rank(symbols, options)
    elif action in ['view', 'show', 'plot']:
        cli_fundflow_view(symbols, options)
    else:
        print('\nError: invalid action: ', action)



# -*- coding: utf-8; py-indent-offset:4 -*-

import requests
import datetime
import time
import pandas as pd
import io

from ..utils import symbol_yahoo_style

# headers and params used to bypass NASDAQ's anti-scraping mechanism in function __exchange2df
yahoo_headers = {
    'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36',
    #'origin': 'https://finance.yahoo.com',
    #'referer': 'https://finance.yahoo.com/',
    #'accept-language': 'en-US,en;q=0.9',
}

#
# Data source:
# https://query1.finance.yahoo.com/v7/finance/download/0700.HK?period1=1594019968&period2=1625555968&interval=1d&events=history&includeAdjustedClose=true
#
def download_us_stock_daily( symbol, start= None, end= None, interval= '1d' ):
    if start is None:
        start = '2010-01-01'

    if isinstance(start, datetime.datetime):
        start = int(time.mktime(start.timetuple()))
    else:
        start = int(time.mktime(time.strptime(str(start), '%Y-%m-%d')))
    if end is None:
        end = int(time.time())
    elif isinstance(end, datetime.datetime):
        end = int(time.mktime(end.timetuple()))
    else:
        end = int(time.mktime(time.strptime(str(end), '%Y-%m-%d')))

    url_template = 'https://query1.finance.yahoo.com/v7/finance/download/{}?period1={}&period2={}&interval={}&events=history&includeAdjustedClose=true'
    url = url_template.format(symbol, start, end, interval)
    #print(url)

    print('\rfetching history data {} ...'.format(symbol), end = '', flush = True)
    r = requests.get(url, headers= yahoo_headers)
    print('')

    if 'Will be right back' in r.text:
        raise RuntimeError("*** YAHOO! FINANCE IS CURRENTLY DOWN! ***.")

    df = pd.read_csv(io.StringIO(r.text))
    df.columns = ['date', 'open', 'high', 'low', 'close', 'adj_close', 'volume']
    df['factor'] = df['adj_close'] / df['close']
    df['date'] = pd.to_datetime(df['date'])
    df = df.set_index('date', drop= True)
    return df

def download_us_index_daily( symbol ):
    return download_us_stock_daily(symbol)

#
# Data source:
# https://query1.finance.yahoo.com/v7/finance/quote?symbols=^GSPC,AAPL,0700.HK,000300.SS
#
def download_us_stock_quote(symbols, verbose = False):
    url_template = 'https://query1.finance.yahoo.com/v7/finance/quote?symbols={}'
    url = url_template.format(','.join(symbols))

    print('\rfetching quote data ...', end = '', flush = True)
    r = requests.get(url, headers= yahoo_headers)
    print('\r', end = '', flush = True)

    if 'Will be right back' in r.text:
        raise RuntimeError("*** YAHOO! FINANCE IS CURRENTLY DOWN! ***.")
    data = r.json()['quoteResponse']['result']
    return pd.DataFrame(data, columns=data[0].keys()) if (len(data) > 0) else pd.DataFrame()

#
# NOTICE: yahoo quote delay 15 min than real market time, don't rely it on realtime trade
#
def download_us_stock_spot(symbols, verbose = False):
    if len(symbols) > 100:
        df = pd.DataFrame()
        for i in range(0, len(symbols), 100):
            page = symbols[i:i+100]
            page_df = download_us_stock_spot(page)
            df = df.append(page_df, ignore_index=True)
        return df
    else:
        df = download_us_stock_quote(symbols, verbose= verbose)
        date = pd.to_datetime(df['regularMarketTime'] + df['gmtOffSetMilliseconds'], unit='s')
        df['date'] = date.dt.normalize()
        df['time'] = date.dt.strftime('%H:%M:%S')
        selected_columns = {
            'symbol': 'symbol',
            'shortName': 'name',
            'regularMarketOpen': 'open',
            'regularMarketPreviousClose': 'prevclose',
            'regularMarketPrice': 'close',
            'regularMarketDayHigh': 'high',
            'regularMarketDayLow': 'low',
            'regularMarketVolume': 'volume',
            'date': 'date',
            'time': 'time',
            'market': 'market',
        }
        df = df[ list(selected_columns.keys()) ].rename(columns= selected_columns)
        return df

def download_world_index_list(param= None, verbose = False):
    symbols = '''^BSESN
^NSEI
^DJI
^IXIC
^N225
^HSI
^AXJO
^TWII
^STI
000001.SS
399001.SZ
^JKSE
^KS11
^GSPC
^DJI
^AORD
^KLSE
^XAX
^RUT
^VIX
^GSPTSE
^FTSE
^GDAXI
^FCHI
^STOXX50E
^N100
^BFX
IMOEX.ME
^BVSP
^MXX
^IPSA
^MERV
^TA125.TA
^CASE30
^JN0U.JO
^NZ50'''.split('\n')
    df = download_us_stock_quote(symbols)
    df = df[['symbol','shortName','exchange','market','currency']]
    df = df.rename(columns={'shortName':'name'})
    return df

#
# Data source:
# https://query1.finance.yahoo.com/v7/finance/spark?symbols=%5EAORD&range=1d&interval=5m&indicators=close&includeTimestamps=false&includePrePost=false&corsDomain=in.finance.yahoo.com&.tsrc=finance
#
def download_us_stock_spark(symbol, verbose= False):
    pass



# -*- coding: utf-8 -*-

from .basic import CROSS
from .ma import SMA
from ..core.indicator_signal import register_signal_indicator

# --------------------------------------------
def ENE(close, length = 25, m1 = 6, m2 = 6):
    ma = SMA(close, length)
    up = ma * (1 + m1/100.0)
    dn = ma * (1 - m2/100.0)
    ene = (up + dn) / 2.0
    return ene, up, dn




# -*- coding: utf-8; py-indent-offset:4 -*-

from .nasdaq import *
from .yahoo import *
from .sina import *
from .akshare_sina import *
from .akshare_tencent import *
from .akshare_legu import *
from .akshare_eastmoney import *



# -*- coding: utf-8; py-indent-offset:4 -*-

from .data_cache import *
from .stock_market import *
from .indicator_signal import *
from .order_cost import *
from .stock import *
from .portfolio import *
from .strategy import *
from .agent_simulated import *
from .agent_human import *
from .fund import *
from .trader import *
from .push_base import *
from .push_email import *
from .push_master import *
from .lang import *
from .backtest import *
from .conf import *



# -*- coding: utf-8 -*-

from ..core.lang import add_lang

add_lang('zh', {
    'buy': '买入',
    'sell': '卖出',
    'stop loss': '止损',
    'stop earn': '止盈',
    'signal': '信号',
    'You can buy': '建议买入',
    'You can sell': '建议卖出',
    'price': '价格',
    'volume': '成交量',
    'trade': '交易',
    'position': '持仓',
    'return(%)': '收益率(%)',
    'position(%)': '仓位(%)',
    'drawdown(%)': '回撤(%)',
    'quantitative backtesting': '量化回测',
})



# -*- coding: utf-8; py-indent-offset:4 -*-

import pandas as pd
import mplfinance as mpf
import matplotlib.pyplot as plt
from cycler import cycler

from .indicator_signal import gen_indicator_signal, signal_to_long, get_all_signal_indicators
from .lang import get_lang, LANG

class Stock:
    symbol = ''
    name = ''
    daily_df = None
    shares = 0
    cost = 0.0
    price = 0.0
    sellable = True

    def __init__(self, symbol, name = '', daily_df = None):
        self.symbol = symbol
        self.name = name
        self.daily_df = daily_df
        self.shares = 0
        self.cost = 0.0
        self.price = 0.0

    def calc_cum_return(self, df, signal, mix= False, order_cost = None):
        # Notice!!! Important !!!
        #
        # when calc signal, we used the close price of today,
        # to avoid "future data or function", it cannot be used for today's trading
        # the signal only impact next day, so here shift(1) to next day
        #
        # only for MFFI, we should catch the signal and buy/sell immediately
        #
        if signal.name == 'mffi':
            long_pos = signal_to_long(signal, time_factor=0.50)
        else:
            long_pos = signal_to_long(signal).shift(1).fillna(0)

        act_return = long_pos * df['daily_return']

        if mix:
            df['trade'] = signal.astype(int)
            df['long_pos'] = long_pos

        # use a column start with "." to store the performance data
        if order_cost is not None:
            buy_cost = signal.copy()
            buy_cost[ buy_cost < 0 ] = 0
            sell_cost = signal.copy()
            sell_cost[ sell_cost > 0 ] = 0
            buy_cost = buy_cost * (order_cost.open_commission)
            sell_cost = (-1) * sell_cost * (order_cost.close_commission + order_cost.close_tax)
            cost = (buy_cost + sell_cost)
        else:
            cost = 0
        cum_return = (1 + act_return - cost).cumprod() -1
        return cum_return.fillna(0)

    def add_indicator(self, indicators, mix = False, inplace = False, order_cost = None):
        # if only one indicator, we auto show trade & position
        if len(indicators) == 1:
            mix = True

        df = self.daily_df
        df['daily_return'] = df.close.pct_change().fillna(0)
        if mix:
            signal = gen_indicator_signal(df, indicators, inplace=inplace)
            if len(indicators) == 1:
                signal.rename(indicators[0], inplace= True)
            df['return.'] = self.calc_cum_return(df, signal, mix= True, order_cost= order_cost)
        else:
            for k in indicators:
                signal = gen_indicator_signal(df, [k], inplace=inplace).rename(k)
                df[k + '.'] = self.calc_cum_return(df, signal, mix= False, order_cost= order_cost)

    def rank_indicator(self, by = 'overall'):
        df = self.daily_df.copy()

        # fill NaN with 0
        df.fillna(0, inplace=True)

        # sort the indicators by overall performance
        table = []
        for k in df.columns:
            if k.endswith('.'):
                v = df[k]
                final = v.iloc[-1]
                overall = sum(v.tolist())
                table.append([k, final, overall])
        rank_df = pd.DataFrame(table, columns=['indicator', 'final', 'overall'])
        rank_df.sort_values(by=by, ascending=False, inplace=True)
        rank_df.reset_index(inplace=True, drop=True)

        return rank_df

    # Visualization
    # --------------------------------------------------------------------------------
    def plot(self, out_file = None):
        # China market candle color, red for up, green for down
        up_red = (get_lang() == 'zh')
        long_color = 'red' if up_red else 'green'
        short_color = 'green' if up_red else 'red'
        mc = mpf.make_marketcolors(
            up = long_color,
            down = short_color,
            edge = 'i',
            wick = 'i',
            volume = 'in',
            inherit = True
        )
        style = mpf.make_mpf_style(
            base_mpl_style = "ggplot",
            marketcolors = mc,
            rc = {
                # to correctly display Chinese text
                # download TTF files:
                # https://raw.githubusercontent.com/StellarCN/scp_zh/master/fonts/SimHei.ttf
                # https://raw.githubusercontent.com/StellarCN/scp_zh/master/fonts/SimKai.ttf
                # https://raw.githubusercontent.com/StellarCN/scp_zh/master/fonts/SimSun.ttf
                # copy to: /usr/local/lib/python3.9/site-packages/matplotlib/mpl-data/fonts/ttf/
                'font.sans-serif': ['SimHei'], # Chinese font
                'font.family': 'sans-serif',
                'axes.unicode_minus': False,
                'axes.prop_cycle': cycler(color=['dodgerblue', 'deeppink', 'navy', 'teal', 'maroon', 'darkorange', 'indigo']),
                'lines.linewidth': 0.75,
            }
        )

        # we have panel 0 for main, panel 1 for volume by default
        next_panel = 2
        more_plot = []

        df = self.daily_df

        linestyles = dict(
            type = 'line',
            width = 1
        )
        panel_ratios = [1, 0.3]
        keys = [
            'ma5', 'ma10', 'ma20', 'ma30', 'ma60',
            'ema5', 'ema10', 'ema20', 'ema30', 'ema60',
            'hma5', 'hma10', 'hma20', 'hma30', 'hma60',
        ]
        cols = list(set(keys) & set(df.columns))
        if len(cols) > 0:
            more_plot.append(mpf.make_addplot(df[cols], panel=0, **linestyles))

        if 'mffi' in df.columns:
            mffi = df['mffi']
            trade_color = [long_color if v >= 0 else short_color for v in mffi]
            more_plot.append(mpf.make_addplot(mffi, type='bar', panel=next_panel, color=trade_color, ylabel='MFFI'))
            next_panel = next_panel +1
            panel_ratios.append(0.8)

        if 'macd_hist' in df.columns:
            macd = df.macd_hist
            macd_color = [long_color if v >= 0 else short_color for v in macd]
            more_plot.append(mpf.make_addplot(macd, type='bar', width=0.4, panel=next_panel, color=macd_color, ylabel='MACD'))
            cols = []
            for k in ['macd_dif', 'macd_dea']:
                if k in df.columns:
                    cols.append(k)
            if len(cols) > 0:
                more_plot.append(mpf.make_addplot(df[cols], panel=next_panel, secondary_y=False, **linestyles))
            next_panel = next_panel +1
            panel_ratios.append(0.3)

        indicators = get_all_signal_indicators()
        for key, values in indicators.items():
            if key in ['ma', 'macd', 'mffi']:
                continue
            ind_label = values['label']
            ind_cols = values['cols']

            if ind_cols[0] in df.columns:
                cols = list(set(ind_cols) & set(df.columns))
                more_plot.append(mpf.make_addplot(df[cols], panel=next_panel, ylabel=ind_label, secondary_y=False, **linestyles))
                next_panel = next_panel +1
                panel_ratios.append(0.3)

        if 'trade' in df.columns:
            # as the signal calculated with close price of today,
            # to avoid "future data", we can only trade before market close, or morning of next day
            trade_action = df['trade']

            trade_color = [long_color if v >= 0 else short_color for v in trade_action]
            more_plot.append(mpf.make_addplot(trade_action, type='bar', panel=next_panel, color=trade_color, ylabel=LANG('trade')))
            next_panel = next_panel +1
            panel_ratios.append(0.3)

        if 'long_pos' in df.columns:
            price_change = df['close'] - df['open']
            pos_color = [long_color if v >= 0 else short_color for v in price_change]
            long_pos = df['long_pos']
            more_plot.append(mpf.make_addplot(long_pos, type='bar', panel=next_panel, color=pos_color, ylabel=LANG('position')))
            next_panel = next_panel +1
            panel_ratios.append(0.3)

        # find the columns started with "." as performance data
        ret_cols = []
        for k in df.columns:
            if k.endswith('.'):
                ret_cols.append(k)
        if len(ret_cols) > 0:
            price_change_percent = (df['close']/df['close'].iloc[0]-1) * 100
            more_plot.append(mpf.make_addplot(price_change_percent, panel=next_panel, color='b', ylabel=LANG('return(%)'), **linestyles))
            more_plot.append(mpf.make_addplot(df[ret_cols] * 100, panel=next_panel, secondary_y=False, **linestyles))
            last_panel = next_panel
            next_panel = next_panel +1
            panel_ratios.append(1.0 if (len(ret_cols)>1) else 1.0)

        print(df)

        kwargs = dict(
            title = self.symbol + ' - ' + self.name,
            type = "candle",
            volume = True,
            datetime_format = '%Y-%m-%d',
            ylabel = LANG("price"),
            ylabel_lower = LANG("volume"),
            figratio = (10,6),
            figscale = 1.0,
            panel_ratios = panel_ratios,
            tight_layout = True,
            style = style,
            addplot = more_plot,
            returnfig = True,
            warn_too_much_data = 244 * 5,
        )
        if out_file is not None:
            kwargs[ 'savefig' ] = out_file

        fig, axes = mpf.plot(df[['open','close','high','low','volume']], **kwargs)

        # matplotlib subplot does not support legend, we do it by ourselves
        if(len(ret_cols) > 0):
            ret_cols.insert(0, 'price')
            axes[ last_panel * 2].legend(ret_cols, loc = 'best')

        plt.show()



# -*- coding: utf-8; py-indent-offset:4 -*-

from ..utils import dict_from_config_items
from .conf import get_hiquant_conf

class OrderCost:
    close_tax = 0.001
    open_commission = 0.0003
    close_commission = 0.0003
    min_commission = 5.0

    def __init__(self, close_tax = 0.001, open_commission = 0.0003, close_commission = 0.0003, min_commission = 5.0):
        self.close_tax = close_tax
        self.open_commission = open_commission
        self.close_commision = close_commission
        self.min_commission = min_commission

def get_order_cost(global_config = None):
    order_cost_conf = None
    if (global_config is not None) and ('order_cost' in global_config.sections()):
        order_cost_conf = dict_from_config_items(global_config.items('order_cost'))
    else:
        config = get_hiquant_conf()
        order_cost_conf = dict_from_config_items(config.items('order_cost'))

    if order_cost_conf is not None:
        return OrderCost(
            float(order_cost_conf['close_tax']),
            float(order_cost_conf['open_commission']),
            float(order_cost_conf['close_commission']),
            float(order_cost_conf['min_commission']),
        )
    else:
        return OrderCost(0.001, 0.0003, 0.0003, 5.0)



# -*- coding: utf-8 -*-

import pandas as pd
import talib as tl
from talib import abstract

from .basic import SIGN

def is_percent(x: int or float) -> bool:
    if isinstance(x, (int, float)):
        return x is not None and x >= 0 and x <= 100
    return False

def INCREASING(close, length=1, strict=False, asint=True, percent=None, drift=1):
    # Validate Arguments
    length = int(length) if length and length > 0 else 1
    strict = strict if isinstance(strict, bool) else False
    asint = asint if isinstance(asint, bool) else True
    percent = float(percent) if is_percent(percent) else False

    if close is None: return

    # Calculate Result
    close_ = (1 + 0.01 * percent) * close if percent else close
    if strict:
        # Returns value as float64? Have to cast to bool
        increasing = close > close_.shift(drift)
        for x in range(3, length + 1):
            increasing = increasing & (close.shift(x - (drift + 1)) > close_.shift(x - drift))

        increasing.fillna(0, inplace=True)
        increasing = increasing.astype(bool)
    else:
        increasing = close_.diff(length) > 0

    if asint:
        increasing = increasing.astype(int)

    return increasing

def DECREASING(close, length=1, strict=False, asint=True, percent=0, drift=1):
    length = int(length) if length and length > 0 else 1
    strict = strict if isinstance(strict, bool) else False
    asint = asint if isinstance(asint, bool) else True
    percent = float(percent) if percent else False

    # Calculate Result
    close_ = (1 - 0.01 * percent) * close if percent else close
    if strict:
        # Returns value as float64? Have to cast to bool
        decreasing = close < close_.shift(drift)
        for x in range(3, length + 1):
            decreasing = decreasing & (close.shift(x - (drift + 1)) < close_.shift(x - drift))

        decreasing.fillna(0, inplace=True)
        decreasing = decreasing.astype(bool)
    else:
        decreasing = close_.diff(length) < 0

    if asint:
        decreasing = decreasing.astype(int)

    return decreasing

positive_patterns = [
    'CDLSTALLEDPATTERN',
    'CDLINNECK',
    'CDLXSIDEGAP3METHODS',
    'CDLGAPSIDESIDEWHITE',
]

negative_patterns = [
    'CDLADVANCEBLOCK',
    'CDLHAMMER',
    'CDLINVERTEDHAMMER',
    'CDLGRAVESTONEDOJI',
    'CDLEVENINGSTAR',
    'CDLEVENINGDOJISTAR',
]

def PATTERN_SIGNAL(price):
    signals = pd.Series([0] * price.shape[0], index=price.index)
    all_patterns = tl.get_function_groups()['Pattern Recognition']

    for name in positive_patterns:
        if name in all_patterns:
            func = abstract.Function(name)
            pattern_signals = func(price)
            signals += pattern_signals

    for name in negative_patterns:
        if name in all_patterns:
            func = abstract.Function(name)
            pattern_signals = func(price)
            signals -= pattern_signals

    return SIGN(signals)



# -*- coding: utf-8 -*-

from .basic import MIN, MAX, NON_ZERO, CROSS
from .ma import RMA
from ..core.indicator_signal import register_signal_indicator

# KDJ index, by George Lane
# --------------------------------------------
def KDJ(close, high, low, length = 9, signal = 3):
    lowest_low = MIN(low, length)
    highest_high = MAX(high, length)
    rsv = (close - lowest_low) / NON_ZERO(highest_high - lowest_low) * 100.0
    k = RMA(rsv, length=signal)
    d = RMA(k, length=signal)
    j = 3 * k - 2 * d
    return k, d, j

def signal_kdj(df, inplace=False, length=9, signal=3):
    k, d, j = KDJ(df.close, df.high, df.low, length=length, signal=signal)
    if inplace:
        df['kdj_k'], df['kdj_d'], df['kdj_j'] = k, d, j
    cross = CROSS(k, d)
    buy_signal = (cross > 0) | (k < 20)
    sell_signal = (cross < 0) | (k > 80)
    return buy_signal.astype(int) - sell_signal.astype(int)

register_signal_indicator('kdj', signal_kdj, ['kdj_k', 'kdj_d', 'kdj_j'], 'KDJ', 'obos')



# -*- coding: utf-8; py-indent-offset:4 -*-

import datetime as dt
import pandas as pd

from ..utils import datetime_today
from .data_cache import get_all_symbol_name, get_cn_stock_fund_flow_rank, \
    get_daily, get_stock_fund_flow_daily, \
    get_stock_spot, \
    get_index_daily

class Market:
    all_symbol_name = {}
    watching_symbols = []

    symbol_daily = {}
    symbol_daily_adjusted = {}

    verbose = False

    date_start = None
    date_end = None
    current_date = None
    current_time = None

    biz_time = None
    biz_hour = (11.5 - 9.5) + (15.0 - 13.0)
    sell_rule = 'T+1'

    is_testing = False

    def __init__(self, start, end):
        self.all_symbol_name = get_all_symbol_name()
        self.watching_symbols = []
        self.symbol_daily = {}
        self.symbol_daily_adjusted = {}

        self.date_start = start
        self.date_end = end
        self.current_date = self.current_time = start
        self.last_spot_time = dt.datetime.now() - dt.timedelta(minutes =10)

        self.set_business_time(morning= ['9:30', '11:30'], afternoon= ['13:00', '15:00'])
        self.set_sell_rule(sell_rule= 'T+1')

    def set_sell_rule(self, sell_rule = 'T+1'):
        self.sell_rule = sell_rule

    def set_business_time(self, morning= ['9:30', '11:30'], afternoon= ['13:00', '15:00']):
        business_time = morning + afternoon

        biz_time = []
        for k in business_time:
            time_only = dt.datetime.strptime(k, '%H:%M')
            biz_time.append(time_only.hour + time_only.minute / 60.0)
        self.biz_time = biz_time

        morning_start, morning_end, afternoon_start, afternoon_end = self.biz_time
        self.biz_hour = (morning_end - morning_start) + (afternoon_end - afternoon_start)

    def set_verbose(self, verbose = True):
        self.verbose = verbose

    # for testing purpose
    def set_date_range(self, start, end):
        self.date_start = start
        self.date_end = end
        self.current_date = self.current_time = start

    def is_open(self) -> bool:
        if self.is_testing:
            return True

        # business hour:
        # monday to friday, 9:30 ~ 11:30, 13:00 ~ 15:00
        now = dt.datetime.now()
        wday = now.weekday() +1
        if (wday >= 1) and (wday <= 5):
            hour = now.hour + now.minute / 60.0
            morning_start, morning_end, afternoon_start, afternoon_end = self.biz_time
            if (hour >= morning_start) and (hour < morning_end):
                return True
            if (hour >= afternoon_start) and (hour < afternoon_end):
                return True

        return False

    def load_history_price(self, symbol):
        if symbol.startswith('sh') or symbol.startswith('sz'):
            df = get_index_daily( symbol )
            self.symbol_daily[ symbol ] = df
            self.symbol_daily_adjusted[ symbol ] = df
        else:
            # OHCLV
            df = get_daily( symbol )

            # fund flow
            fund_df = get_stock_fund_flow_daily( symbol )
            if fund_df is not None:
                df['main_fund'] = fund_df['main_fund']
                df['main_pct'] = fund_df['main_pct']
                df = df.fillna(0)
            self.symbol_daily[ symbol ] = df

            if 'factor' in df.columns:
                df = df.copy()
                for k in ['open', 'close', 'high', 'low']:
                    df[k] = df[k] * df['factor']
            self.symbol_daily_adjusted[ symbol ] = df

    def watch(self, symbols):
        for symbol in symbols:
            if not (symbol in self.watching_symbols):
                self.watching_symbols.append(symbol)
                self.load_history_price(symbol)

    def keep_daily_uptodate(self):
        for symbol in self.watching_symbols:
            self.load_history_price( symbol )

    def update_fundflow_realtime(self, verbose = False) -> pd.DataFrame:
        today = pd.to_datetime( datetime_today() )

        # merge fund flow data into df
        fund_df = get_cn_stock_fund_flow_rank()
        fund_df = fund_df[ fund_df.symbol.isin(self.watching_symbols) ]
        fund_df = fund_df.sort_values(by='main_pct', ascending=False).reset_index(drop= True)
        if verbose:
            print('')
            print(fund_df)

        for i, fund_row in fund_df.iterrows():
            symbol = fund_row['symbol']
            if symbol not in self.watching_symbols:
                continue
            df = self.symbol_daily_adjusted[ symbol ]
            if today in df.index:
                if 'main_pct' in df.columns:
                    df.loc[ today ]['main_pct'] = fund_row['main_pct']
                if 'main_fund' in df.columns:
                    df.loc[ today ]['main_fund'] = fund_row['main_fund']

        return fund_df

    def update_daily_realtime(self, verbose = False) -> bool:
        now = dt.datetime.now()
        data_updated = False

        # if the spot out of date, update for all stocks under watch
        next_spot_time = self.last_spot_time + dt.timedelta(seconds=10)
        if now > next_spot_time:
            spot_df = get_stock_spot(self.watching_symbols, verbose)
            if verbose:
                print('')
                print(spot_df)
            self.last_spot_time = now

            today = pd.to_datetime( datetime_today() )
            for i, spot_row in spot_df.iterrows():
                spot_date = spot_row['date'] if ('date' in spot_row) else today
                if spot_date >= today:
                    data_updated = True

                    symbol = spot_row['symbol']

                    new_row = self.symbol_daily[ symbol ].iloc[-1].copy()
                    for k in ['open', 'high', 'low', 'close', 'volume']:
                        if k in spot_row:
                            new_row[ k ] = spot_row[ k ]
                    self.symbol_daily[ symbol ].loc[spot_date] = new_row

                    if 'factor' in new_row:
                        new_row = new_row.copy()
                        adjust_factor = new_row['factor']
                        for k in ['open', 'high', 'low', 'close']:
                            new_row[ k ] = new_row[ k ] * adjust_factor
                    self.symbol_daily_adjusted[ symbol ].loc[spot_date] = new_row

        self.update_fundflow_realtime(verbose= verbose)

        return data_updated

    def get_index_daily(self, symbol, start = None, end = None, count = None) -> pd.DataFrame:
        if end is None:
            end = self.date_end
        if start is None:
            start = self.date_start - dt.timedelta(days=90)
        df = get_index_daily( symbol )

        if start:
            df = df[ pd.to_datetime(start) : ]
        if end:
            df = df[ : pd.to_datetime(end) ]
        if count:
            df = df.tail(count)

        return df

    def get_daily(self, symbol, start = None, end = None, count = None) -> pd.DataFrame:
        if end is None:
            end = self.date_end
        if start is None:
            start = self.date_start - dt.timedelta(days=90)

        if not symbol in self.watching_symbols:
            self.watch([ symbol ])
            if end and (end >= datetime_today()):
                self.update_daily_realtime()

        df = self.symbol_daily_adjusted[ symbol ]
        if start:
            df = df[ pd.to_datetime(start) : ]
        if end:
            df = df[ : pd.to_datetime(end) ]
        if count:
            df = df.tail(count)

        return df

    def get_spot(self, symbol, date = None) -> pd.Series:
        if date is None:
            date = self.current_date

        df = self.get_daily(symbol, end = date, count = 1)

        if df.shape[0] > 0:
            return df.iloc[-1]
        else:
            return None

    def get_time_factor(self) -> float:
        cur_hour = self.current_time.hour + self.current_time.minute / 60.0

        # convert to valid biz time
        morning_start, morning_end, afternoon_start, afternoon_end = self.biz_time
        cur_hour = min(max(cur_hour, morning_start), afternoon_end)
        if (cur_hour > morning_end) and (cur_hour < afternoon_start):
            cur_hour = afternoon_start

        hour = (cur_hour - morning_start) if (cur_hour <= morning_end) else (cur_hour - afternoon_start + (morning_end - morning_start))
        return hour / self.biz_hour

    def get_price(self, symbol, date = None) -> float:
        if date is None:
            date = self.current_date

        if type(symbol) == list:
            return [self.get_price(sym) for sym in symbol]

        if symbol == 'cash':
            return 1.0

        # adjusted daily
        df = self.get_daily(symbol, end = date, count = 1)

        if df.shape[0] > 0:
            row = df.iloc[-1]
            close_price = row['close']
            if self.current_time > datetime_today():
                return close_price
            else:
                open_price = row['open']
                # assume the price change is function of time, either linear, quadratic, or cubic
                f = 1 - self.get_time_factor()
                price = close_price - (close_price - open_price) * (f * f * f)
                return price
        else:
            print(date, 'not trading day')
            return 0

    def get_real_price(self, symbol, date = None) -> float:
        if date is None:
            date = self.current_date

        if type(symbol) == list:
            return [self.get_real_price(sym) for sym in symbol]

        if symbol == 'cash':
            return 1.0

        if not symbol in self.watching_symbols:
            self.watch([ symbol ])
            today = datetime_today()
            if date and (date >= today):
                self.update_daily_realtime()

        df = self.symbol_daily[ symbol ]
        df = df[:date]

        if df.shape[0] > 0:
            row = df.iloc[-1]
            close_price = row['close']
            if self.current_time > datetime_today():
                return close_price
            else:
                open_price = row['open']
                # assume the price change is function of time, either linear, quadratic, or cubic
                f = 1 - self.get_time_factor()
                price = close_price - (close_price - open_price) * f * f * f
                return price
        else:
            print(date, 'not trading day')
            return 0

    def get_adjust_factor(self, symbol, date = None) -> pd.Series:
        df = self.get_daily(symbol, end = date, count = 1)
        factor =  df['factor'].iloc[-1] if (df.shape[0] > 0) else False
        if not factor:
            print(df)
        return factor

    def get_name(self, symbol) -> str:
        if symbol in self.all_symbol_name:
            return self.all_symbol_name[ symbol ]
        else:
            return ''

    def allow_trading(self, symbol, date = None) -> bool:
        if date is None:
            date = self.current_date
        df = self.symbol_daily_adjusted[ symbol ]
        return pd.to_datetime(date) in df.index

    def get_all_trading_dates(self) -> list:
        date_set = ()
        for symbol, df in self.symbol_daily.items():
            date_set |= set(df.index)
        dates = list(date_set)
        dates.sort()
        return dates

    def get_main_fundflow_rank(self, symbols) -> pd.DataFrame:
        if self.current_date >= datetime_today():
            # get the fund flow ranking from web for today
            df = self.update_fundflow_realtime()
            df = df[ df['symbol'].isin(symbols) ]
            df = df[['symbol','main_pct', 'main_fund']]
        else:
            # or, merge them from history data
            table = []
            for symbol in symbols:
                if not symbol in self.watching_symbols:
                    self.watch([ symbol ])
                main_pct, main_fund = 0.0, 0.0
                if symbol in self.symbol_daily_adjusted:
                    daily_df = self.symbol_daily_adjusted[ symbol ]
                    the_date = pd.to_datetime(self.current_date)
                    if the_date in daily_df.index:
                        row = daily_df.loc[ the_date ]
                        main_pct, main_fund = row['main_pct'], row['main_fund']
                else:
                    print('symbol not in symbol_daily_adjusted:', symbol)
                row = [symbol, main_pct, main_fund]
                table.append(row)

            df = pd.DataFrame(table, columns=['symbol', 'main_pct', 'main_fund'])

        # now sort by main fundflow percentage
        df = df.sort_values(by='main_pct', ascending= False).reset_index(drop= True)
        return df

    def get_main_fundflow(self, symbol) -> set:
        if symbol in self.symbol_daily_adjusted:
            daily_df = self.symbol_daily_adjusted[ symbol ]
            the_date = pd.to_datetime( self.current_date )
            if the_date in daily_df.index:
                row = daily_df.loc[ the_date ]
                if ('main_pct' in row) and ('main_fund' in row):
                    return (row['main_pct'], row['main_fund'])
        return (0.0, 0.0)
